# -*- python -*-
# $RCSfile: gfxLabelTree.py,v $
# $Revision: 1.31 $
# $Author: reida $
# $Date: 2005/01/14 16:07:19 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# A wrapper for a GTK+ CTree widget whose structure mirrors a given
# LabelTree.  The GfxLabelTree is automatically updated when the
# LabelTree changes.

# Selecting, deselecting, or double clicking a tree entry
# calls the user-supplied callback function.  The callback function
# takes at least two arguments: the name of the signal which triggered
# the callback ('select', 'deselect', or 'doubleclick') and the node
# of the tree that was selected or deselected.  Extra arguments to the
# GfxLabelTree constructor are also passed through to the callback.


from SWIG.common import switchboard
from common import labeltree
from common import debug
from common.IO.GUI import chooser
import string
from gtk import *


class GfxLabelTree:
    def __init__(self, tree, callback=None, *callbackargs, **callbackkwargs):
        debug.mainthreadTest()
        self.tree = tree                # associated LabelTree
        self.callback = callback
        self.callbackargs = callbackargs
        self.callbackkwargs = callbackkwargs
        self.gtk2lt = {}                # maps gtk tree items to LabelTree nodes
        self.lt2gtk = {tree:None}
        self.selectedNode = None        # selected LabelTree node

        self.gtk = GtkCTree(cols=1, tree_col=0, titles=None)
        self.gtk.set_selection_mode(SELECTION_SINGLE)
        for node in tree.nodes:
            self.constructGUI(node, None)

        self.selection_signal = self.gtk.connect("tree-select-row",
                                                 self.selectCB)
        self.deselection_signal = self.gtk.connect("tree-unselect-row",
                                                   self.unselectCB)
        self.gtk.connect("button_press_event", self.buttonpressCB)
        self.doubleclick = 0
        self.gtk.connect("destroy", self.destroyCB)
        self.autoSelect()
        self.gtk.show_all()

        self.sbcallbacks = [
            switchboard.requestCallback((tree, "insert"), self.insertCB),
            switchboard.requestCallback((tree, "delete"), self.deleteCB),
            switchboard.requestCallback((tree, "rename"), self.renameCB)
            ]
        
    def constructGUI(self, labeltreenode, gtktreeparent):
        if labeltreenode.secret():
            return
        gtknode = self.gtk.insert_node(gtktreeparent, None,
                                       [labeltreenode.name],
                                       is_leaf=FALSE, expanded=TRUE)
        self.gtk2lt[gtknode] = labeltreenode
        self.lt2gtk[labeltreenode] = gtknode
        if labeltreenode.nodes:
            for node in labeltreenode.nodes:
                self.constructGUI(node, gtknode)

    def selectCB(self, tree, node, column): # gtk callback
        self.selectedNode = self.gtk2lt[node]
        if self.selectedNode.object is not None:
            if self.callback:
                if self.doubleclick:
                    self.callback("doubleclick", self.selectedNode,
                                  *self.callbackargs, **self.callbackkwargs)
                    self.doubleclick = 0
                else:
                    self.callback("select", self.selectedNode,
                                  *self.callbackargs, **self.callbackkwargs)
        else:
            # Oops. Tried to select a node that doesn't have an object
            # stored.  Since we don't know how to prevent the
            # selection in the first place, we just deselect it
            # immediately.  At this point, self.unselectCB() has
            # already been called for the old selection, so there's no
            # need to call self.callback again.
            self.deselect()
            
    def unselectCB(self, tree, node, column): # gtk callback
        if self.callback:
            selectee = self.gtk2lt[node]
            self.selectedNode = None
            self.callback("deselect", selectee,
                          *self.callbackargs, **self.callbackkwargs)

    def buttonpressCB(self, tree, gdkevent):
        if gdkevent.type == GDK._2BUTTON_PRESS:
            # The imminent call to selectCB() will pass the
            # "doubleclick" signal to the callback function.
            self.doubleclick = 1
        
    # Programmatically select an object stored in the tree, without
    # calling the callback function.
    def selectObject(self, object):
        node = self.tree.reverse_dict[object]
        self.selectNode(node)

    # Programmatically select a LabelTree node, without calling the
    # callback function.
    def selectNode(self, node):
        debug.mainthreadTest()
        gtknode = self.lt2gtk[node]
        self.gtk.signal_handler_block(self.selection_signal)
        self.gtk.select(gtknode)
        self.gtk.signal_handler_unblock(self.selection_signal)
        self.selectedNode = node

    def deselect(self):
        # Check self.tree to suppress deselect signals caused by
        # widget destruction.
        debug.mainthreadTest()
        if self.tree and self.selectedNode:
            gtknode = self.lt2gtk[self.selectedNode]
            self.gtk.signal_handler_block(self.deselection_signal)
            self.gtk.unselect(gtknode)
            self.gtk.signal_handler_unblock(self.deselection_signal)
            self.selectedNode = None

    def getSelection(self):
        if self.selectedNode is not None:
            return self.selectedNode.object

    # Ensure that the given labeltree node is visible, by expanding
    # each node along the path.
    def expandToPath(self, path):
        namelist = string.split(path,':')
        name_sublists = [ namelist[:i] for i in range(1,len(namelist)+1) ]
        pathlist = [ string.join(x,":") for x in name_sublists ]
        for p in pathlist:
            node = self.tree[p]
            self.gtk.expand(self.lt2gtk[node])

    def autoSelect(self):
        if self.selectedNode is None:
            # If the tree has only a single leaf, select it.
            leafpaths = self.tree.leafpaths()
            if len(leafpaths) == 1:
                node = self.tree[leafpaths[0]]
                self.selectNode(node)

    def destroyCB(self, *args):         # gtk callback
        map(switchboard.removeCallback, self.sbcallbacks)
        # clean up possible circular references
        self.tree = None                
        del self.selectedNode
        del self.gtk2lt
        del self.lt2gtk
        del self.callback
        del self.callbackargs
        del self.callbackkwargs

    def destroy(self):
        debug.mainthreadTest()
        self.gtk.destroy()

    # Switchboard callbacks for insertion and deletion.  This allows
    # the GUI to remain synchronized to the dynamic state of the
    # underlying tree.

    def insertCB(self, parent, node):
        # parent and node are LabelTreeNode objects
        # Currently ignores the fact that these lists are nominally ordered.
        gtkparent = self.lt2gtk[parent]
        self.constructGUI(node, gtkparent)
        self.autoSelect()

    def deleteCB(self, node):
        # node is a LabelTreeNode object
        debug.mainthreadTest()
        gtknode = self.lt2gtk[node]
        self.gtk.remove_node(gtknode)
        del self.lt2gtk[node]
        del self.gtk2lt[gtknode]
        self.autoSelect()

    def renameCB(self, node):
        debug.mainthreadTest()
        # node is a LabelTreeNode object
        gtknode = self.lt2gtk[node]
        self.gtk.node_set_text(gtknode, 0, node.name)


######################################

# LabelTree representation as a pile of ChooserWidget objects.  Does
# not (yet) update itself when the tree changes.

class LabelTreeChooserWidget:
    def __init__(self, tree, value=None, callback=None, scope=None,
                 condition=None, sort=None):
        debug.mainthreadTest()
        self.callback = callback
        self.tree = tree
        self.condition = condition
        self.sort = sort

        if scope:
            scope.addWidget(self)

        self.gtk = GtkVBox()
        self.widgets = []
        if value is not None:
            self.set_value(value)
        else:
            self.set_value(tree.numberOneChild())
    def show(self):
        self.gtk.show_all()

    def set_value(self, object):
        debug.mainthreadTest()
        path = labeltree.makePath(self.tree.objpath(object))
        depth = len(path)
        for widget in self.widgets:
            widget.destroy()
        self.widgets = []
        tree = self.tree
        depth = 0
        for name in path:
            names = [node.name for node in tree.nodes]
            widget = chooser.ChooserWidget(names, callback=self.chooserCB,
                                           callbackargs=(depth,))
            depth += 1
            widget.set_state(name)
            self.widgets.append(widget)
            self.gtk.add(widget.gtk, expand=0, fill=0)
            tree = tree[name]
        self.gtk.show_all()
    def chooserCB(self, gtkobj, name, depth):
        leadpath = [self.widgets[d].get_value() for d in range(depth)]
        subtree = self.tree[leadpath+[name]]
        self.set_value(subtree.numberOneChild())
        if self.callback:
            self.callback()

    def get_value(self):
        path = [widget.get_value() for widget in self.widgets]
        if path:
            return self.tree[path].object
        
        

