# -*- python -*-
# $RCSfile: progressbar_.py,v $
# $Revision: 1.31 $
# $Author: redwing $
# $Date: 2004/11/08 16:35:06 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


from common import debug
from common.IO import threadmanager


from SWIG.common import switchboard


_progressbar_options = [
    "continuous",
    "active",
    "discrete"
    ]


## A ProgressBar is an object created by the worker in threaded mode
## and placed in the threadManager. These objects constitute the base
## for instantiating a representation (text or graphic), when an
## activityViewer window opens. See (***)

## When the activityViewer window is closed, the ProgressBar
## representations are destroyed, but the ProgressBar objects (the
## "core" of the progress bars) remain. 

## If an activityViewer window is already open and a new ProgresBar
## is created, a switchboard call message is issued and the activityViewer
## is automatically updated.

## The progress bar representation hierarchy is the following:
##
## A ProgressBarRepresentation is the base of all the graphic, text,
## acoustic, tickle, or otherwise UI that the user will experience (maybe
## one day in stereo).
##
## (***) The correct representation is instantiated by calling the function
## set_ProgressBarRepresentation, which is over-written in graphic mode.
##
## The TextProgressBar is a subclass of the ProgressBarRepresentation and
## has all the functionality to interact with the user in text mode.
##
## The GUIProgressBar is s subclass of the ProgressBarRepresentation and
## has all the functionality to interact with the user in graphics mode.


## IMPORTANT NOTE on the ProgressObject class:

## A ProgressObject is NOT derived from the Progress class
## because when one is instantiated within a thread it checks
## that the C++ counter-part is (or not is) already in the static
## list of progress objects. The ProgressObject object possesses a
## reference to the correct Progress object, which is shared between
## the C++ and the Python levels.
## The functionality of the ProgressObject is inhereted through the
## use of a __getattr__ function.

class ProgressObject :
    def __init__(self, type = "continuous", name = None):
        self.progressbar = getProgress()
        if type in _progressbar_options: ## check validity
            self.type = type ## assign progressbar type
        else:
            raise AttributeError("progress bar option not valid")
        self.Worker = None
        self.name = name
        self.defunct = 0
        ## After progress object is created, pass it to threadManager
        self.register_to_Thread_Manager()
        
    def __getattr__(self, attrname):
        ## Please see comment above
        return getattr(self.progressbar, attrname)
    
    
## These are the basic functions for a ProgressObject
    def get_name(self):
        return self.name
    
    def set_name(self, name):
        self.name = name
        
    def registerWorker(self, wrkr):
        self.Worker = wrkr
        
        
    def unregisterWorker(self):
        self.Worker = None

    def register_to_Thread_Manager(self):
        threadmanager.threadManager.newProgressBar(self)
        
    def get_type(self):
        return self.type
        
    def get_Progress(self):
        return self.progressbar

    
#########################################################################
#########################################################################
## The following is the basic class for creating a progress bar
## representation. Each representation has different API for drawing
## the progress bar.
#########################################################################
#########################################################################
    
class ProgressBarRepresentation:
    def __init__(self, progressBar):
        self.progressbar = progressBar
        self.type = self.progressbar.get_type()
        self.name = self.progressbar.get_name()
        self.talk = 1
    def get_name(self):
        return self.progressbar.get_name()
    def set_progress(self, improvement):
        return self.progressbar.set_progress(improvement)
        
#########################################################################
#########################################################################
## The following describes the progress bar representation for a text
## progress bar. In this case, the progress bar object returnd a string
## whose length and contents depends on the current state of the thread.
#########################################################################
#########################################################################
        
class TextProgressBar(ProgressBarRepresentation):
    def __init__(self, progressBar):
        ProgressBarRepresentation.__init__(self, progressBar)
        self.bar_length = 10
        self.window_width = 80
        self.bar_repr = None
        self.total_bar = " "

        ## shuttle mode progress bar
        self.direction = 1
        self.counter = 0

    def __getattr__(self, attrname):
        return getattr(self.progressbar, attrname)
        

    def bar_size(self):
        possible_sz =  self.window_width- len(self.get_message())- 5
        if possible_sz > 0:
            return possible_sz
        return 5
    
    def set_window_size(self, szz):
        self.window_width = szz
        
    ## functions to draw different types of progress bars
    def progress_active(self):
        actual_bar = ""
        ## print id(self)
        ## print "counter",self.counter
        self.bar_length = self.bar_size()
        ## print "direction %d barlength %d"%(self.direction, self.bar_length) 
        for i in range(self.bar_length):
            if  i==self.counter:
                actual_bar += "*"
            else:
                actual_bar += " "
        if self.counter > self.bar_length or self.counter <0 :
            self.direction *= -1
        self.counter  += 3*self.direction
        ## print ">>" + actual_bar + "<<"
        ## print "counter",self.counter
        ## print actual_bar, self.counter, self.direction
        return actual_bar

    def progress_incremental(self, progress):
        actual_bar = ""
        self.bar_length = self.bar_size()
        for i in range(self.bar_length):
            if i<=self.bar_length*progress:
                actual_bar += "*"
            else:
                actual_bar +=" "
        return actual_bar
    
    def buildprogressbar(self):
        ## Updates text progress bar.
        ## For activity mode, the passed value 
        ## is irrelevant. Valid values range 
        ## between zero and one.
        string_fraction = ""
        if self.type == "active":
            string_fraction = self.progress_active()
        elif self.type == "continuous" or "discrete":
            string_fraction = self.progress_incremental(self.get_progress())
        self.total_bar = self.progressbar.get_message() + " |" + string_fraction + "| " + "\r"
        return self.total_bar

    

   

def set_ProgressBarRepresentation(pbar): ## returns a text progress bar
    ## argument is a ProgressObject.. This function is redefined if the
    ## GUI is loaded.
    return TextProgressBar(pbar)
    



