# -*- python -*-
# $RCSfile: meshparamwidgets.py,v $
# $Revision: 1.46 $
# $Author: reida $
# $Date: 2005/02/08 15:06:26 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

# ParameterWidgets for things that depend on a Mesh, such as Fields,
# Fluxes, and Equations defined on the Mesh.

from SWIG.common import switchboard
from SWIG.engine import equation
from SWIG.engine import field
from SWIG.engine import flux
from SWIG.engine import planarity
from common import debug
from common.IO.GUI import chooser
from common.IO.GUI import parameterwidgets
from common.IO.GUI import whowidget
from engine import component
from engine import skeletoncontext
from engine.IO import meshparameters
from gtk import *
import engine.mesh
import string

class MeshParamWidget(parameterwidgets.ParameterWidget):
    # Base class for a widget that displays and allows choices from a
    # list of things from a Mesh (eg, defined Fields or active
    # Equations).  The meshfunc constructor argument is a function
    # that returns the list of things to display.  It's called with
    # the mesh as its first argument, so it can be a Mesh member
    # function.  self.chooser is the ChooserWidget that displays the
    # items returned by meshfunc.

    def __init__(self, param, meshfunc, scope):
        self.meshfunc = meshfunc
        self.chooser = chooser.ChooserWidget([], callback=self.chooserCB)
        parameterwidgets.ParameterWidget.__init__(self, self.chooser.gtk, scope)
        self.meshwidget = scope.findWidget(
            lambda w: isinstance(w, whowidget.WhoWidget)
            and w.whoclass is engine.mesh.meshes)
        self.sbcallbacks = [switchboard.requestCallback(self.meshwidget,
                                                        self.update)]
        self.update(interactive=0)
        if param.value is not None:
            self.set_value(param.value)
        
    def update(self, interactive):
        mesh = self.getMesh()
        self.vals = {None:None}
        if mesh is not None:
            namelist = []
            for obj in self.meshfunc(mesh): 
                self.vals[obj.name()] = obj
                namelist.append(obj.name())
            self.chooser.update(namelist)
            self.widgetChanged(len(namelist) > 0, interactive)
        else:
            self.chooser.update([])
            self.widgetChanged(0, interactive)
    def chooserCB(self, *args):
        self.widgetChanged(1, interactive=1)
    def getMesh(self):
        meshname = self.meshwidget.get_value()
        try:
            return engine.mesh.meshes[meshname]
        except KeyError:
            return None
    def get_value(self):
        return self.vals[self.chooser.get_value()]
    def set_value(self, obj):
        self.chooser.set_state(obj.name())
        self.widgetChanged(1, interactive=0)
        
    def cleanUp(self):
        self.meshwidget = None
        map(switchboard.removeCallback, self.sbcallbacks)
        parameterwidgets.ParameterWidget.cleanUp(self)


# Widgets for quantities to which Invariants can be calculated should
# be subclasses of InvariandWidget.

class InvariandWidget: pass

# Widgets for quantities for which components can be extracted should
# be subclasses of IndexableWidget.

class IndexableWidget: pass

#############################

def fieldlister(mesh):                  # meshfunc for FieldParameterWidgets
    # list fields and out-of-plane fields, if defined.\
    # TODO: This can be called with an unresolvable proxy if there
    # are no meshes.
    compoundfields = engine.mesh.Mesh.all_fields(mesh)
    flds = []
    for fld in compoundfields:
        flds.append(fld)
        if not mesh.getObject().in_plane(fld):
            flds.append(fld.out_of_plane())
    return flds
        
class FieldParameterWidget(MeshParamWidget, InvariandWidget, IndexableWidget):
    def __init__(self, param, scope):
        if param.outofplane:
            flist = fieldlister
        else:
            flist = engine.mesh.Mesh.all_fields
        MeshParamWidget.__init__(self, param, flist, scope)


def _makeFieldWidget(param, scope):
    return FieldParameterWidget(param, scope)

meshparameters.FieldParameter.makeWidget = _makeFieldWidget

#############################

class FluxParameterWidget(MeshParamWidget, InvariandWidget, IndexableWidget):
    def __init__(self, param, scope):
        MeshParamWidget.__init__(self, param,
                                 engine.mesh.Mesh.all_fluxes, scope)

def _makeFluxWidget(param, scope):
    return FluxParameterWidget(param, scope)

meshparameters.FluxParameter.makeWidget = _makeFluxWidget

#############################

class EquationParameterWidget(MeshParamWidget, IndexableWidget):
    def __init__(self, param, scope):
        MeshParamWidget.__init__(self, param, engine.mesh.Mesh.all_equations,
                                 scope)
        self.sbcallbacks.append(
            switchboard.requestCallback('equation activated', self.newEqn))
    def newEqn(self, mesh, equation, active):
        if mesh is self.getMesh():
            self.update()

def _makeEquationWidget(param, scope):
    return EquationParameterWidget(param, scope)

meshparameters.EquationParameter.makeWidget = _makeEquationWidget

#############################################
#############################################

class FieldIndexParameterWidget(parameterwidgets.ParameterWidget):
    def __init__(self, param, scope):
        debug.mainthreadTest()
        self.chooser = chooser.ChooserWidget([], callback=self.chooserCB)
        box = GtkVBox()
        box.add(self.chooser.gtk, expand=0, fill=0)
        parameterwidgets.ParameterWidget.__init__(self, box, scope)
        self.fieldwidget = scope.findWidget(
            lambda w: isinstance(w, IndexableWidget))
        self.sbcallback = switchboard.requestCallback(self.fieldwidget,
                                                      self.fieldCB)
        self.notapplicable = GtkLabel('(Not Applicable)')
        self.notapplicable.set_alignment(0.0, 0.5)
        box.add(self.notapplicable, expand=0, fill=0)
        self.nIndices = 0
        self.update()
        if param.value is not None and param.value != '':
            self.chooser.set_state(param.value)
        self.widgetChanged(1, interactive=0)
    def chooserCB(self, *args):
        self.widgetChanged(1, interactive=1)
    def fieldCB(self, interactive):
        self.update()
        self.widgetChanged(1, interactive)
    def update(self):                   # field has changed
        itlist = []
        self.nIndices = 0
        field = self.fieldwidget.get_value()
        if field is not None:
            iterator = field.iterator(planarity.ALL_INDICES)
            while not iterator.end():
                self.nIndices += 1
                it = iterator.cloneIndex()
                itrepr = it.shortrepr()
                itlist.append(itrepr)
                iterator.next()
        self.chooser.update(itlist)
        self.show()
    def show(self):
        self.gtk.show()
        if self.nIndices > 1:
            self.chooser.gtk.show()
            self.notapplicable.hide()
        else:
            self.chooser.gtk.hide()
            self.notapplicable.show()
    def get_value(self):
        val = self.chooser.get_value()
        if val is None:
            return ''
        return val
    def cleanUp(self):
        parameterwidgets.ParameterWidget.cleanUp(self)
        switchboard.removeCallback(self.sbcallback)

def _makeFieldIndexParameterWidget(param, scope):
    return FieldIndexParameterWidget(param, scope)

meshparameters.FieldIndexParameter.makeWidget = _makeFieldIndexParameterWidget


############################################################

# Special widget for mesh boundary parameters.  As it turns out, these
# params take strings, not boundary objects, so we override a
# significant fraction of the foregoing machinery, which is really
# designed for parameters which take the actual object.

class MeshBoundaryParamWidget(MeshParamWidget):
    def __init__(self, param, scope):
        MeshParamWidget.__init__(self, param,
                                 engine.mesh.Mesh.boundaryNames,
                                 scope)
        self.sbcallbacks.append(
            switchboard.requestCallback('mesh boundaries changed',
                                        self.newBdys) )
    def newBdys(self, mesh):
        if mesh is self.getMesh():
            self.update(interactive=0)
    def update(self, interactive):
        mesh = self.getMesh()
        if mesh is not None:
            namelist = self.meshfunc(mesh)
            self.chooser.update(namelist)
            if namelist: # If list has nonzero length, widget is valid.
                self.widgetChanged(1, interactive)
        else:
            self.chooser.update([])
            self.widgetChanged(0, interactive)
    def set_value(self, value):
        self.chooser.set_state(value)
        self.widgetChanged(1, interactive=0)
    def get_value(self):
        return self.chooser.get_value()

def _makeBoundaryWidget(param, scope):
    return MeshBoundaryParamWidget(param, scope)

meshparameters.MeshBoundaryParameter.makeWidget = _makeBoundaryWidget




# Special cases for nontrivial edge and point boundaries.


        
class MeshEdgeBdyParamWidget(MeshBoundaryParamWidget):
    def __init__(self, param, scope):
        MeshParamWidget.__init__(self, param,
                                  engine.mesh.Mesh.edgeBoundaryNames,
                                  scope)
        self.sbcallbacks.append(
            switchboard.requestCallback('mesh boundaries changed',
                                        self.newBdys) )

def _makeEdgeBdyWidget(param, scope):
    return MeshEdgeBdyParamWidget(param, scope)

meshparameters.MeshEdgeBdyParameter.makeWidget = _makeEdgeBdyWidget




class MeshPointBdyParamWidget(MeshBoundaryParamWidget):
    def __init__(self, param, scope):
        MeshParamWidget.__init__(self, param,
                                  engine.mesh.Mesh.pointBoundaryNames,
                                  scope)
        self.sbcallbacks.append(
            switchboard.requestCallback('mesh boundaries changed',
                                        self.newBdys) )


def _makePointBdyWidget(param, scope):
    return MeshPointBdyParamWidget(param, scope)

meshparameters.MeshPointBdyParameter.makeWidget = _makePointBdyWidget
