# -*- python -*-
# $RCSfile: skeletonIPC.py,v $
# $Revision: 1.42 $
# $Author: redwing $
# $Date: 2004/07/21 21:01:19 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

import math, sys
from common import debug
from common import object_id
from common import primitives
from common import utils
from common.IO import binarydata
from common.IO import mainmenu
from common.IO import oofmenu
from common.IO import parallelmainmenu
from common.IO import parameter
from common.IO import socket2me
from engine import skeleton
from engine import skeletonboundary
from engine import skeletonsegment
from engine import skeletonselectionmod
from engine.IO import skeletonIO
from SWIG.common import mpitools
from SWIG.common import ooferror
from SWIG.common import switchboard


## General definitions
OOF = mainmenu.OOF
_rank = mpitools.rank()
_size = mpitools.size()


## Containers for boundaries
class EdgeBoundaries:
    ## edge boundaries are referred to by name
    def __init__(self, btm, top, rgt, lft):
        self.edgeboundaries = {}
        self.edgeboundaries[btm.get_name()] = btm
        self.edgeboundaries[top.get_name()] = top
        self.edgeboundaries[rgt.get_name()] = rgt
        self.edgeboundaries[lft.get_name()] = lft
    def getEdge(self, name):
        return edgeboundaries[name]
    def addEdge(self, edg):
        self.edgeboundaries[edg.get_name()] = edg
    
class PointBoundaries:
    ## point boundaries are also referred to by name
    def __init__(self, tl, tr, bl, br):
        self.pointboundaries = {}
        self.pointboundaries[tl.get_name()] = tl
        self.pointboundaries[tr.get_name()] = tr
        self.pointboundaries[bl.get_name()] = bl
        self.pointboundaries[br.get_name()] = br
    def getPointBoundary(self, name):
        return self.pointboundaries[name]
    def addPointBoundary(self, pbdy):
        self.pointboundaries[pbdy.get_name()] = pbdy

## functions to find free and marked elements
def boundary_free(el):
    ## This function was borrowed from the
    ## engine.skeletonselectionmod.ExpandElementSelection class
    ## It finds a boundary element that does not have a rank set
    for n in el.nodes:
        for e in n.neighborElements():
            if e.get_ownership() <0:
                return 1
    return 0

def boundary_marked(el, p_number):
    ## This function iterates over
    ## the neighbor elements of el, and if
    ## one with a different owner is found
    ## return 1
    for n in el.nodes:
        for e in n.neighborElements():
            if e.get_ownership() != p_number:
                return 1
    return 0

## functions to set global and local indices
def dictTolist(dictionary):
    ## list = []
    ##for key in dictionary.keys():
    ##    list.append(dictionary[key])
    ##return list
    return dictionary.keys()

def listTodict(list):
    dict = {}
    for el in list:
        dict[el] = el
    return dict

def set_global_index(dictionary, obj_list):
    ## rank has been set at this point!
    _gindex = 0
    for obj in obj_list:
        obj.ID.set_index(_gindex)
        dictionary[(_gindex, obj.ID.get_rank())] = obj
        _gindex += 1
    return

def set_index( dictionary, obj_list, p_owner):
    ## rank has been set at this point!
    _index = 0
    for obj in obj_list:
        if obj.ID.get_rank() == p_owner:
            obj.ID.set_index(_index)
            dictionary[(_index, p_owner)] = obj
            _index += 1
    return


## functions to set rank in nodes and segments
def select_nodes(element, proc):
    for nd in element.nodes:
        if nd.get_ownership()<0:
            nd.set_ownership(proc) ## C++
            nd.ID.set_rank(proc) ## python


def select_segments(skeleton, element, p_owner):
    segments = element.getSegments(skeleton)
    for sgmnt in segments:
        if sgmnt.get_ownership() <0:
            sgmnt.set_ownership(p_owner) ## python
            sgmnt.ID.set_rank(p_owner) ## python



##functions to set processor number in rectangle
def interior(xmin, xmax, ymin, ymax, n):
    ## returns 1 if n (a node) is between xmin--xmax and ymin--ymax
    if n.position().x < xmax and n.position().x > xmin and \
           n.position().y < ymax and n.position().y > ymin:
        return 1
    return 0

def element_inside(xmin, xmax, ymin, ymax, element):
    x_pos = 0
    y_pos = 0
    num_elements = 0
    for node in element.nodes:
        x_pos += node.position().x
        y_pos += node.position().y
        num_elements += 1.0
    x_pos /= num_elements
    y_pos /= num_elements
    if x_pos < xmax and x_pos > xmin and \
           y_pos < ymax and y_pos > ymin:
        return 1
    return 0

def select_rectangle(skeleton, ul, lr, p_owner):
    ## this function returns a list of elements whose owner is p_owner
    
    ## skeleton is a real skeleton
    xmin = min(ul.x, lr.x)
    xmax = max(ul.x, lr.x)

    ymin = min(ul.y, lr.y)
    ymax = max(ul.y, lr.y)

    for e in skeleton.elements:
        sflag = 0
        if element_inside(xmin, xmax, ymin, ymax, e):
                sflag = 1
        else:
            for n in e.nodes:
                if interior(xmin, xmax, ymin, ymax, n):
                    sflag = 1
        if sflag == 1:
            if e.get_ownership() < 0:
                e.set_ownership(p_owner)
                e.ID.set_rank(p_owner)
                select_nodes(e, p_owner) ## ID is set inside
                select_segments(skeleton, e, p_owner) ## ID is set inside

                

def add_from_neighbor_processor(elist, p_number):
    ## Here, a list of boundary elements that
    ## does not belong to the current processor is
    ## added to the list. This list will be passed to
    ## the p_number.
    new_els = []
    for el in elist:
        for n in el.nodes:
            for e in n.neighborElements():
                if e.get_ownership() is None:
                    raise ooferror.ErrPyProgrammingError("element does not have a rank!")
                if e.get_ownership() != p_number:
                    if e not in new_els:
                        new_els.append(e)
    ## add only new elements
    ## debug.fmsg("will add ", len(new_els), " boundary elements")
    for nel in new_els:
        if nel not in elist:
            elist.append(nel)
    return elist
   


def get_elements(skeleton, p_number):
    elist = []
    for e in skeleton.elements:
        if e.get_ownership() == p_number:
            elist.append(e)
    return elist

def _rectangular_marking(skel):
    ## skel is a REAL skeleton
    global _size
    n = int(math.pow(_size, 0.5)) ## assume we have n^2 processors
    width  = skel.MS.size()[0]*1.0
    height = skel.MS.size()[1]*1.0
    dx = width/n  
    dy = height/n
    ul = primitives.Point(0,0)
    lr = primitives.Point(dx, dy)
    for p_owner in range(_size):
        select_rectangle(skel, ul, lr, p_owner)
        ## selects elements and nodes
        if ul.x <= dx*(n-2):
            ul.x += dx
        else: ## if ur has reached an edge, then move to the next block
            ul.x = 0   
        if lr.x <= dx*(n-1):
            lr.x += dx
        else: ## if lr has reached an edge, then move to the next block
            lr.x = dx
            ul.y += dy
            lr.y += dy
        
def _greedy_selector(self, target_size, seed_element, p_owner):
    ## this function returns a list of elements whose owner is p_owner
    seed_element.set_ownership(p_owner)
    select_nodes(seed_element, p_owner)
    selection = []
    blements = []
    selection.append(seed_element) ## initially, the seed is the whole boundary
    _marked = len(selection)
    while _marked <= target_size :
        for tentative in selection:
            if boundary_free(tentative):
                ## if boundary element does not have an processor
                ## save it to a list, so it receives an owner
                belements.append(el)
            
        if len(belements) >0:
            ## if we found boundary elements, then
            ## set the processors owner and
            ## add them to the selection list
            for bel in belements:
                for n in bel.nodes:
                    for e in n.neighborElements():
                        if e.get_ownership() <0:
                            e.set_ownership(p_owner)
                            select_nodes(e, p_owner)
                            select_segments(skeleton, e, p_owner)
                            selection.append(e)
                        if len(selection) == target_size:
                            ## if the maximum size of the list
                            ## has been reached, then return
                            ## the list
                            return selection
            belements = [] ## clear list for next pass
        else:
                ## if there are no boundaries left (only the seed?)
                ## then return what he got so far
                return selection
            
        _marked = len(selection)
    return selection

def _mark_for_parallelization(self):
    _rectangular_marking(self)
    ## _greedy_selector(self) ## for later...

skeleton.Skeleton.mark_for_parallelization = _mark_for_parallelization


## OOF.LoadData.IPC.Skeleton
smenu = parallelmainmenu.ipcmenu.addItem(oofmenu.OOFMenuItem('Skeleton', secret=1, no_log=1))



def get_nodes(edict):
    node_list = []
    for e in edict:
        for n in e.nodes:
            if n not in node_list:
                node_list.append(n)
    return node_list

def getpoint_boundaries(skeleton, node_list):
    ## this function assumes that the point boundaries are parallelized
    ## right when the skeleton is constructed and there is no other
    ## type of point boundaries, but the defined ones in the Quad or Tri
    ## skeletons
    btmlft = skeleton.getPointBoundary('bottomleft')
    btmrgt = skeleton.getPointBoundary('bottomright')
    toplft = skeleton.getPointBoundary('topleft')
    toprgt = skeleton.getPointBoundary('topright')
    
    btmlft_rdcd = skeletonboundary.SkeletonPointBoundary('bottomleft')
    btmrgt_rdcd = skeletonboundary.SkeletonPointBoundary('bottomright')
    toplft_rdcd = skeletonboundary.SkeletonPointBoundary('topleft')
    toprgt_rdcd = skeletonboundary.SkeletonPointBoundary('topright')
    for nd in node_list.keys():
        if btmlft.has_node(nd) :
            ## debug.fmsg("btmlft", nd)
            btmlft_rdcd.addNode(nd)
        elif btmrgt.has_node(nd):
            ## debug.fmsg("btmrght", nd)
            btmrgt_rdcd.addNode(nd)
        elif toplft.has_node(nd):
            ## debug.fmsg("toplft", nd)
            toplft_rdcd.addNode(nd)
        elif toprgt.has_node(nd) :
            ## debug.fmsg("toprgt", nd)
            toprgt_rdcd.addNode(nd)
    ## some point boundaries could be empty
    return PointBoundaries(toplft_rdcd, toprgt_rdcd, btmlft_rdcd, btmrgt_rdcd)

def getedge_boundaries(skeleton, node_list):
    btm = skeleton.getEdgeBoundary('bottom')
    top = skeleton.getEdgeBoundary('top')
    rgt = skeleton.getEdgeBoundary('right')
    lft = skeleton.getEdgeBoundary('left')
    
    btm_rdcd= skeletonboundary.SkeletonEdgeBoundary('bottom')
    top_rdcd= skeletonboundary.SkeletonEdgeBoundary('top')
    rgt_rdcd= skeletonboundary.SkeletonEdgeBoundary('right')
    lft_rdcd= skeletonboundary.SkeletonEdgeBoundary('left')

    ## bottom edge
    for edge in btm.edges:
        nd1, nd2 = edge.get_nodes()
        if nd1 in node_list.keys():
            if nd2 in node_list.keys():
                ## debug.fmsg()
                btm_rdcd.addEdge(edge)
    ## top edge
    for edge in top.edges:
        nd1, nd2 = edge.get_nodes()
        if nd1 in node_list.keys():
            if nd2 in node_list.keys():
                ## debug.fmsg()
                top_rdcd.addEdge(edge)

    ## right edge
    for edge in rgt.edges:
        nd1, nd2 = edge.get_nodes()
        if nd1 in node_list.keys():
            if nd2 in node_list.keys():
                ## debug.fmsg()
                rgt_rdcd.addEdge(edge)

    ## left edge
    for edge in lft.edges:
        nd1, nd2 = edge.get_nodes()
        if nd1 in node_list.keys():
            if nd2 in node_list.keys():
                ## debug.fmsg()
                lft_rdcd.addEdge(edge)

    return EdgeBoundaries(btm_rdcd, top_rdcd, rgt_rdcd, lft_rdcd)


def get_segments(skeleton, node_list, p_number):
    ## returns a list of segments that does NOT
    ## belong to processor p_number
    segment_list = []
    skel_segments = skeleton.segments
    for sgmnt in skel_segments.keys():
        nd0, nd1 = sgmnt
        if nd0 in node_list:
            if nd1 in node_list:
                if skel_segments[sgmnt].get_ownership() != p_number:
                    segment_list.append(skel_segments[sgmnt])
    return segment_list




def _send_skeleton_elements(skeletoncontext, element_list, nodes_list, skeleton_segments, destination):
    ## DO NOT PUT LOCKS HERE.
    ## What follows is based on engine.IO.skeletonIO.writeSkeleton()
    ## The locks are already set in  _parallel_init()
    
    # Nodes are written and read in the order in which they're stored
    # in the Skeleton's nodes list.  This is not necessaryly the same as the
    # local machines' internal index, so we need to create a dictionary of
    # indices in the nodes list so that the elements' node tuples are
    # correct.
    nodedict = {} ## local: only for rebuilding the elements on the other side
    c = 0
    
    for node in nodes_list:
        nodedict[node] = c
        c += 1
    skeleton = skeletoncontext.getObject()
    mpitools.send_string(skeletoncontext.name(), destination) ## send skeleton name
    mpitools.send_string(skeleton.MS.name(), destination) ## send microstructure name
    
    node_positions = []
    _ids = []
    for nd in nodes_list:
        node_positions.append(nd.position().x)
        node_positions.append(nd.position().y)
        _ids.append(nd.ID.get_index()) ## send global index
        _ids.append(nd.ID.get_rank()) ## send rank
    mpitools.send_vector_of_doubles(node_positions, destination) ## send node positions
    ## debug.fmsg("sending nodes=", node_positions)
    ## debug.fmsg("node ids=", _ids)
    mpitools.send_vector_of_ints(_ids, destination) ## send node ids
    # Define elements using the node indices in nodedict.
    el_list = []
    _ids = [] 
    for el in element_list:
        el_list.append(len(el.nodes)) ## signal number of nodes
        for nd in el.nodes:
            el_list.append(nodedict[nd]) ## send temporary index
        _ids.append(el.ID.get_index()) ## send global index
        _ids.append(el.ID.get_rank()) ## send global rank
    mpitools.send_vector_of_ints(el_list, destination) ## send node key positions for each element
    ## debug.fmsg("element ids=", _ids)
    mpitools.send_vector_of_ints(_ids, destination) ## send element ids
    el_list = []
    _ids = []

    ## create temporary lists of point boundaries, edge boundaries
    pointboundaries = getpoint_boundaries(skeleton, nodedict)
    edge_boundaries = getedge_boundaries(skeleton, nodedict)
    ## point_boundaries and edge_boundaries hold a PointBoundary and Edgeboundary object, respectively
    
    # Point boundaries
    _int_node_list = []
    mpitools.send_int(len(pointboundaries.pointboundaries.items()), destination)
    for pbname, pbdy in pointboundaries.pointboundaries.items():
        mpitools.send_string(pbname, destination) ## point boundary name
        for node in pbdy.nodes:
            _int_node_list.append(nodedict[node])
        ## debug.fmsg("point boundary ", pbname,_int_node_list)
        mpitools.send_vector_of_ints(_int_node_list, destination)
        exterior = 0
        if isinstance(pbdy, skeletonboundary.ExteriorSkeletonPointBoundary):
            exterior = 1
        mpitools.send_int(exterior, destination)
        _int_node_list = []
        
    # Edge boundaries
    _edge_list = []
    mpitools.send_int(len(edge_boundaries.edgeboundaries.items()), destination)
    for ebname, ebdy in edge_boundaries.edgeboundaries.items():
        mpitools.send_string(ebname, destination) ## edge boundary name
        for edge in ebdy.edges:
            for nd in edge.get_nodes():
                ## debug.fmsg(ebname, edge.get_nodes(), edge)
                _edge_list.append(nodedict[nd])
            
        mpitools.send_vector_of_ints(_edge_list, destination)
        exterior = 0
        if isinstance(ebdy, skeletonboundary.ExteriorSkeletonEdgeBoundary):
            exterior = 1
        mpitools.send_int(exterior, destination)
        _edge_list = []
        
    # Skeleton Element segments
    ## Send only those segments that do NOT belong
    ## to the local processor
    for edg in skeleton_segments:
        node0, node1 = edg.nodes()
        try:
            _edge_list.append(nodedict[node0])
            _edge_list.append(nodedict[node1])
            _ids.append(edg.ID.get_index())
            _ids.append(edg.ID.get_rank())
        except Exception:
            raise
    mpitools.send_vector_of_ints(_edge_list, destination)
    mpitools.send_vector_of_ints(_ids, destination)



def _receive_skeleton(source):
    ## DO NOT PUT LOCKS HERE.
    ## The locks are already set in  _parallel_init()
    skelname = mpitools.receive_string(source)
    mic = mpitools.receive_string(source)
    skel = skeleton.newEmptySkeleton(skelname, mic)
    node_list = mpitools.receive_vector_of_doubles(source) ## holds x and y positions
    ## print "node_list", node_list
    ## sys.stdout.flush()
    _ids_list = mpitools.receive_vector_of_ints(source)
    _index = 0
    _size = len(node_list)
    points = []
    _ids = []
    while _index<_size: ## this can't be the best solution...
        points.append((node_list[_index],node_list[_index+1]))
        _ids.append((_ids_list[_index],_ids_list[_index+1]))
        _index += 2

    for _index in range(len(points)):
        node = points[_index]
        ## debug.fmsg(node)
        nd = skel.newNode(*node) ## add nodes to newly created skeleton
        ## debug.fmsg("mobility x", nd.movable_x(), " mobility y", nd.movable_y())
        gindex, rnk = _ids[_index]
        ## debug.fmsg(gindex, rnk)
        nd.ID.set_id(rnk, gindex)
        ## also set the node dictionary
        skel.node_dictionary[(gindex, rnk)] = nd
    ## debug.fmsg("node id keys", skel.node_dictionary.keys())
    # Define elements, using the node indices in nodedict.
    _ids_list = []
    _ids = []
    points = []
    el_node_list = mpitools.receive_vector_of_ints(source) ## receive node index for each element
    _ids_list = mpitools.receive_vector_of_ints(source) ## receive element ids
    _index = 0
    _size = len(el_node_list)
    ##debug.fmsg(el_node_list)
    ## debug.fmsg("element id list", _ids_list)
    while _index < _size:
        el_num_nodes = el_node_list[_index]
        temp_list = []
        for next_item in range(el_num_nodes):
            temp_list.append(el_node_list[_index + next_item +1])
        skel.loadElement(*temp_list)
        _index += el_num_nodes +1 ## size of previous "packet"
    _index = 0
    _size = len(_ids_list)
    _ids = []
    while _index<_size: 
        _ids.append((_ids_list[_index],_ids_list[_index+1]))
        ## skel.elements[_index].ID.set_id(_ids_list[_index+1], _ids_list[_index])
        _index += 2
    _index = 0
    for _index in range(len(skel.elements)):
        gindex, rnk = _ids[_index] 
        skel.elements[_index].ID.set_id(rnk, gindex)
        skel.element_dictionary[(gindex, rnk)] = skel.elements[_index]
    ## debug.fmsg("element id keys", skel.element_dictionary.keys())
    el_list = []
    _ids = []
    # Point boundaries
    num_of_pboundaries = mpitools.receive_int(source)
    for pbdry_num in range(num_of_pboundaries):
        pbname = mpitools.receive_string(source)
        _node_list = mpitools.receive_vector_of_ints(source)
        pbtype = mpitools.receive_int(source)
        temp_n_list = []
        for i in _node_list:
            temp_n_list.append(skel.nodes[i])
        pbdry = skel.getPointBoundary(pbname, pbtype)
        for node in temp_n_list:
            pbdry.addNode(node)
        
    # Edge boundaries
    num_of_eboundaries = mpitools.receive_int(source)
    for ebdry_index in range(num_of_eboundaries):
        ebname = mpitools.receive_string(source)
        _edge_list = mpitools.receive_vector_of_ints(source)
        ebtype = mpitools.receive_int(source)
        _index = 0
        _size = len(_edge_list)
        edgeBoundary = skel.getEdgeBoundary(ebname, ebtype)
        while _index < _size:
            nd0 = skel.nodes[_edge_list[_index]]
            nd1 = skel.nodes[_edge_list[_index+1]]
            segment = skel.segments[
                        utils.canonicalorder(nd0, nd1)]
            edge = skeletonsegment.SkeletonEdge(segment)
            edge.set_direction(nd0, nd1)
            edgeBoundary.addEdge(edge)
            _index += 2
    ## Skeleton Element segments
    ## Receive only those segments that do NOT belong
    ## to the local processor
    _seg_list = mpitools.receive_vector_of_ints(source)
    _ids = mpitools.receive_vector_of_ints(source)
    _index = 0
    _size = len(_seg_list)
    _segments = []
    while _index < _size:
        nd0 = skel.nodes[_seg_list[_index]]
        nd1 = skel.nodes[_seg_list[_index+1]]
        segment = skel.getSegment(nd0, nd1)
        _id_index = _ids[_index]
        _id_rank = _ids[_index+1]
        segment.ID.set_id(_id_rank, _id_index)
        skel.segment_dictionary[(_id_index, _id_rank)] = segment
        _index += 2

    return skel


## OOF.LoadData.IPC.Skeleton.Initialize
## Partitions the skeleton and sends the pieces to the back-end
def _parallel_init(menuitem, skeleton):
    ## debug.fmsg()
    ## skeleton is the name of the skeleton
    global _rank
    import engine.skeletoncontext
    ## front-end work
    if _rank == 0:
        ## get skeleton from skeleton context
        skelwho = engine.skeletoncontext.skeletonContexts[skeleton]
        skel = skelwho.getObject()
        engine.skeletoncontext.skeletonContexts[skeleton].begin_reading()
        try:
            ## marks skeleton into subregions that
            ## will become subskeletons
            skel.mark_for_parallelization()
            ## TODO LATER: _parallel_init should also use
            ## an argument that selects a skeleton partitioning
            ## algorithm ("rectangle", "greedy", "first moment of inertia axis", etc.)
            set_global_index(skel.node_dictionary, skel.nodes)
            set_global_index(skel.element_dictionary, skel.elements)
            for p_number in range(_size):
                ## create sub-skeletons and send them to the back-end
                new_elist = get_elements(skel, p_number)
                ##  debug.fmsg("number of elements", len(new_elist))
                ## add extra row of elements
                add_from_neighbor_processor(new_elist, p_number)
                ##debug.fmsg("number of elements", len(new_elist))
                ## get expanded dictionary of nodes
                node_list= get_nodes(new_elist)
                
                skeleton_segments= get_segments(skel, node_list, p_number)
                set_index(skel.segment_dictionary, skeleton_segments, p_number)
                if _rank != p_number:
                    ## do not send the list to yourself
                    _send_skeleton_elements(skelwho, new_elist, node_list, skeleton_segments, p_number)
                ## sends marked elements plus an extra bit that is "shared" among
                ## processors, so that future communication is minimized.
        finally:
            engine.skeletoncontext.skeletonContexts[skeleton].end_reading()
        
            
    ## back-end part of the work
    else:
        back_end_skel = _receive_skeleton(0) 
        ## engine.skeletoncontext.skeletonContexts.add(skeleton, back_end_skel)
        skelcontext = engine.skeletoncontext.skeletonContexts[skeleton]
        switchboard.notify(('who changed', 'Skeleton'), skelcontext)

        


smenu.addItem(oofmenu.OOFMenuItem('Initialize', callback = _parallel_init, secret=1, no_log=1, threadable = oofmenu.PARALLEL_THREADABLE, params=[parameter.StringParameter('skeleton')] ))




def _push_skel(menuitem, skeleton, id):
    ## skeleton is the skeleton FULL name
    ## id is the index in the parallel object manager
    global _rank
    ## front-end work
    if _rank == 0:
        pass
    ## back-end part of the work
    else:
        ## debug.fmsg("skeleton=", skeleton, " id=", id)
        ## back-end looks-up the skeleton in the object manager
        ## and pushes it into the skeletoncontext
        from common import parallel_object_manager
        from engine import skeletoncontext
        skel_obj_copy = parallel_object_manager.parallelObjectManager.getObject(id)
        skelwho = skeletoncontext.skeletonContexts[skeleton]
        skelwho.reserve()
        try:
            skelwho.begin_writing()
            try:
                skelwho.pushModification(skel_obj_copy)
            finally:
                skelwho.end_writing()
        finally:
            skelwho.cancel_reservation()
            parallel_object_manager.parallelObjectManager.delete(id)
        
smenu.addItem(oofmenu.OOFMenuItem('Push', callback = _push_skel, secret=1, no_log=1, threadable = oofmenu.PARALLEL_THREADABLE, params=[parameter.StringParameter('skeleton'), parameter.IntParameter('id')] ))

def _undo_skeleton(menuitem, skeleton):
    ## image is the image name
    global _rank
    ## front-end work
    if _rank == 0:
        pass
        
    ## back-end part of the work
    else:
        ## debug.fmsg()
        from engine import skeletoncontext
        skeletonwho = skeletoncontext.skeletonContexts[skeleton]
        skeletonwho.begin_writing()
        try:
            skeletonwho.undoModification()            
        finally:
            skeletonwho.end_writing()
smenu.addItem(oofmenu.OOFMenuItem('Undo', callback = _undo_skeleton, secret=1, no_log=1, threadable = oofmenu.PARALLEL_THREADABLE, params=[parameter.StringParameter('skeleton')] ))


def _redo_skeleton(menuitem, skeleton):
    ## image is the image name
    global _rank
    ## front-end work
    if _rank == 0:
        pass
        
    ## back-end part of the work
    else:
        ## debug.fmsg()
        ## back-end reconstructs the image and pushes it into the imagecontext
        from engine import skeletoncontext
        skeletonwho = skeletoncontext.skeletonContexts[skeleton]
        skeletonwho.begin_writing()
        try:
            skeletonwho.redoModification()
        finally:
            skeletonwho.end_writing()
smenu.addItem(oofmenu.OOFMenuItem('Redo', callback = _redo_skeleton, secret=1, no_log=1, threadable = oofmenu.PARALLEL_THREADABLE, params=[parameter.StringParameter('skeleton')] ))
