// -*- C++ -*-
// $RCSfile: cskeleton.C,v $
// $Revision: 1.50 $
// $Author: reida $
// $Date: 2005/03/01 18:32:52 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/cmicrostructure.h"
#include "common/pixelsetboundary.h"
#include "engine/ooferror.h"
#include "engine/cskeleton.h"
#include "engine/mastercoord.h"
#include "engine/twomatrix.h"
#include <iostream>
#include <math.h>
#include <algorithm>
#include <assert.h>

const unsigned char CSkeletonNode::xmovable_ = 1;
const unsigned char CSkeletonNode::ymovable_ = 2;
const unsigned char CSkeletonNode::unpinned_ = 4;

void CSkeletonNode::setMobilityX(bool mob) {
  if(mob)
    mobility |= xmovable_;
  else
    mobility &= ~xmovable_;
}

void CSkeletonNode::setMobilityY(bool mob) {
  if(mob)
    mobility |= ymovable_;
  else
    mobility &= ~ymovable_;
}

void CSkeletonNode::setPinned(bool pin) {
  if(pin)
    mobility &= ~unpinned_;
  else
    mobility |= unpinned_;
}

// moveTo and moveBy should not be called directly, except by a
// Skeleton or DeputySkeleton object.  Everyone else should use
// Skeleton.moveNodeTo, etc.

bool CSkeletonNode::moveTo(const Coord *pos) {
  lastposition_ = position_;
  lastmoved = nodemoved;
  position_ = *pos;
  if(!movable_x()) 
    position_(0) = lastposition_(0);
  if(!movable_y())
    position_(1) = lastposition_(1);
  if(position_ != lastposition_) {
    ++nodemoved;
    return true;
  }
  return false;
};

void CSkeletonNode::unconstrainedMoveTo(const Coord *pos) {
  lastposition_ = position_;
  lastmoved = nodemoved;
  position_ = *pos;
  ++nodemoved;
}

bool CSkeletonNode::moveBy(const Coord *shift) {
  lastposition_ = position_;
  lastmoved = nodemoved;
  Coord target = position_ + *shift;
  position_ = target;
  if(!movable_x())
    position_(0) = lastposition_(0);
  if(!movable_y())
    position_(1) = lastposition_(1);
  if(position_ != lastposition_) {
    ++nodemoved;
    return true;
  }
  return false;
}

void CSkeletonNode::moveBack() {
  position_ = lastposition_;
  nodemoved = lastmoved;
}

// Can this node be moved to and merge with another node?
bool CSkeletonNode::canMergeWith(const CSkeletonNode *other) const {
  bool xmovable = movable_x();
  bool ymovable = movable_y();
  if(!xmovable && !ymovable)
    return false;
  else if(!xmovable && ymovable)
    return position_(0) == other->position_(0);
  else if(xmovable && !ymovable)
    return position_(1) == other->position_(1);
  return true;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Is pnt inside the triangle (n0, n1, n2)?

static bool insideTriangle(const Coord &n0, const Coord &n1,
			   const Coord &n2, const Coord &pnt)
{
  Coord r0(n0 - pnt);
  Coord r1(n1 - pnt);
  Coord r2(n2 - pnt);
  return r0%r1 >= 0.0 && r1%r2 >= 0.0 && r2%r0 >= 0.0; // cross products
}

static double triangleArea(const Coord &v0, const Coord &v1, const Coord &v2) {
  return 0.5*(v1 - v0)%(v2 - v0);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void CSkeletonElement::replaceNode(int which, CSkeletonNode *replacement) {
  nodes[which] = replacement;
}

std::vector<Coord> *CSkeletonElement::perimeter() const {
  int n = nnodes();
  std::vector<Coord> *pvec = new std::vector<Coord>(n);
  for(int i=0; i<n; i++)
    (*pvec)[i] = nodes[i]->position();
  return pvec;
}

double CSkeletonElement::perimeterLength() const {
  int n = nnodes();
  double p = 0.0;
  for(int i=0; i<n; i++) {
    int j = (i+1)%n;
    Coord r(nodes[i]->position() - nodes[j]->position());
    p += sqrt(r.normsq());
  }
  return p;
}

double CSkeletonElement::edgeLength(int which) const {
  Coord r(nodes[which]->position() - nodes[(which+1)%nnodes()]->position());
  return sqrt(r.normsq());
}

double CSkeletonElement::cosCornerAngle(int which) const {
  int n = nnodes();
  Coord p0 = nodes[which]->position();
  Coord side0 = nodes[(which+n-1)%n]->position() - p0;
  Coord side1 = nodes[(which+1)%n]->position() - p0;
  double cosAngle = dot(side0, side1)/sqrt(norm2(side0)*norm2(side1));
  // Due to an unsuspecting round-off error, this step is needed.
  // (Sometimes cosAngle that should -1 turn out to be -1.000000001)
  if (cosAngle > 1.0)
    return 1.0;
  else if (cosAngle < -1.0)
    return -1.0;
  else
    return cosAngle;
}

Coord CSkeletonElement::center() const {
  int n = nnodes();
  Coord c = nodes[0]->position();
  for(int i=1; i<n; i++)
    c += nodes[i]->position();
  return (1./n)*c;
}

Coord CSkeletonElement::frommaster(MasterCoord *mc, int rotation) const {
  int n = nnodes();
  Coord result(0., 0.);
  for(int i=0; i<n; i++)
    result += shapefun(i, *mc) * nodes[(i+rotation)%n]->position();
  return result;
}

// Returns a vector of doubles.  Each double in this vector is equal
// to the area of the corresponding microstructure category which is
// within this element.  This version uses category boundary
// intersections, but has the same API as other methods of computing
// the categoryArea.

typedef std::multimap<double, PixelBdyIntersection> ElEdgeMap;
typedef std::vector<ElEdgeMap> ElEdge;
typedef std::pair<double, PixelBdyIntersection> ElEdgeDatum;

typedef std::multimap<Coord, PixelBdyIntersection> CoordIsec;
typedef std::pair<Coord, PixelBdyIntersection> CoordIsecDatum;

const std::vector<double> *
CSkeletonElement::new_categoryAreas(CMicrostructure &ms) const {

  // TODO: What to do if the element is illegal?  Assumption of
  // convexity is made below in only handling icounts of zero, one, or
  // two.  This routine should skip illegal elements, incuding
  // negative area elements or nonconvex quadrilaterals.
  int n = ms.nCategories();
  std::vector<double> *result = new std::vector<double>(n);
  bool areas_zero = true;
  int nn = nodes.size();

  const std::vector<PixelSetBoundary> &bdys = ms.getCategoryBdys();

//   for(int i=0; i<bdys.size(); i++) {
//     std::cerr << "Boundary of category " << i << std::endl;
//     std::cerr << bdys[i] << std::endl;
//   }
  
#ifdef DEBUG
  assert(bdys.size()==n);
#endif
  
  PixelBdyIntersection pbi;
  std::vector<PixelBdyIntersection> isecs(0);
  isecs.reserve(4);
  
  for(int i=0; i<bdys.size(); i++) {
    double accum_area = 0.0;
    const LoopVec &loops = bdys[i].get_loops();
    
    // Store intersection data -- used to figure out what sections
    // of the element boundaries contribute to the area.  Stored a
    // couple of ways -- firstly by element edge, so it can be
    // easily traversed, and secondly indexed by coordinate, so
    // duplicates can be removed.
    ElEdge eledgedata(nn);
    CoordIsec coordisecs;
    
    for(LoopVec::const_iterator j = loops.begin(); j!=loops.end(); ++j ) {

      std::vector<PixelBdySegment>::const_iterator k = (*j).begin();

      // TODO: Maybe write a version of "interior" that takes a reference?
      // Then we could skip this variable-and-assignment.
      Coord loop_start = (*k).start_pt();
      bool inside = interior( &loop_start );
      
      // Loop over the segments in this loop.
      for(; k!=(*j).end(); ++k) {
	// Now, loop over all the element-bounding segments.
	for(int n=0; n < nn; n++) {  // Not at all confusing...
	  const Coord &p1 = nodes[n]->position();
	  const Coord &p2 = nodes[(n+1)%nn]->position(); // Slow?
	  // std::cerr << "Skeleton edge: " << p1 << ", " << p2 << std::endl;
	  if ((*k).find_intersection(p1, p2, pbi) ) {
	    isecs.push_back(pbi);

	    // Find the length at which this intersection
	    // occurs, and insert it with that key into a multimap.
	    // This generates an ordered list of the intersections
	    // which occur on a particular element edge.
	    double f = (pbi.location-p1).normsq();
	    pbi.set_element_data(n, f);
	    coordisecs.insert(CoordIsecDatum(pbi.location, pbi));

//  	    if (pbi.entry) {
//  	      std::cerr << "Found entry intersection at " << pbi.location 
//  			<< std::endl;
//  	    }
//  	    else { 
//  	      std::cerr << "Found exit intersection at " << pbi.location 
//  			<< std::endl;
//  	    }

	    
	    eledgedata[n].insert(ElEdgeDatum(f, pbi) );
	  }
	}
	int icount = isecs.size();
	// For wholly-interior segments, add to the current set.
	if (icount==0 && inside) {
	  accum_area += (*k).start_pt() % (*k).end_pt();
	}
	else if (icount==1) {
	  if ( isecs[0].entry ) {
	    accum_area += isecs[0].location % (*k).end_pt();
	    inside = true;
	  }
	  else { // isecs[0].entry false, so it's an exit.
	    accum_area += (*k).start_pt() % isecs[0].location;
	    inside = false;
	  }
	}
	else if (icount==2) { // Must be one entry and one exit.
	  if(isecs[0].entry) {
// 	    std::cerr << isecs[0].location << " and " 
// 		      << isecs[1].location << std::endl; 
	    accum_area += isecs[0].location % isecs[1].location;
	  }
	  else {
// 	    std::cerr << isecs[1].location << " and " 
// 	    	      << isecs[0].location << std::endl;
	    accum_area += isecs[1].location % isecs[0].location; 
	  }
	  // All legal elements are convex, so entry/exit must leave
	  // you outside.
	  inside = false;
	} 
	isecs.clear();
	isecs.reserve(4);

      } // End of segments.
    } // End of loops.
      

    // Debugging -- list all intersections.
//     std::cerr << std::endl << "Pre-annihilation, all intersections: " << std::endl;
//     for(int ni = 0; ni<nn; ++ni) {
//       for(ElEdgeMap::iterator there = eledgedata[ni].begin();
// 	  there!=eledgedata[ni].end(); ++there) {
// 	std::cerr << "Intersection: " << (*there).second.location << 
// 	  ((*there).second.entry ? ", Entry" : ", Exit" ) << std::endl;
//       }
//     }
      

    // Annihilate opposite-polarity intersection pairs which occur
    // at the same coordinates.  These are easily found in the
    // coordisecs multimap.
    CoordIsec::iterator x = coordisecs.begin();
    while(x!=coordisecs.end()) {
      Coord key = (*x).first;
      int ct = coordisecs.count(key);
      if (ct>1) {
	CoordIsec::iterator w = x;

	// Set x to the start of the next key, for the next
	// iteration of the outer loop.
	x = coordisecs.upper_bound(key);
	
	// Make a copy of the problem section of the map.  This copy
	// may not actually be required, one could re-retrieve all
	// the elements with the current key each time.
	CoordIsec key_range(w,x); 
	
	bool done = false;
	while(!done) {
	  if (key_range.size() < 2) { done = true; break; }
	  CoordIsec::iterator furst = key_range.begin();
	  CoordIsec::iterator other = furst; other++;
	  
	  // Assume we are done.  If an opposite-polarity pair is
	  // found, done will be set to false, and another round
	  // will be started.  If no opposite-polarity pairs are
	  // found, then we actually really are done.
	  done = true;
	  while(other!=key_range.end()) {
	    if ( (*furst).second.entry != (*other).second.entry ) {
	      
	      // Delete these two intersections from their
	      // respective eledgedata maps.  The key is given by
	      // get_element_fraction(), and the map to look in is
	      // given by get_element_node_index.  The
	      // fraction-based retrieval may not be unique!
	      int ni = (*furst).second.get_element_node_index();
	      double fr = (*furst).second.get_element_fraction();
	      int count = eledgedata[ni].count(fr);
	      if (count==1) {
		eledgedata[ni].erase(fr);
	      }
	      else { // Find the right one.
		ElEdgeMap::iterator emi = eledgedata[ni].lower_bound(fr);
		for(int i=0; i<count; ++emi, ++i) {
		  if (((*emi).second.location==(*furst).second.location) && 
		      ((*emi).second.entry==(*furst).second.entry) ) {
		    eledgedata[ni].erase(emi);
		    break;
		  }
		  // Might be diagnostically useful to know if you
		  // broke out or not.
		}
	      }
	      
	      ni = (*other).second.get_element_node_index();
	      fr = (*other).second.get_element_fraction();
	      count = eledgedata[ni].count(fr);
	      if (count==1) {
		eledgedata[ni].erase(fr);
	      }
	      else {
		ElEdgeMap::iterator emi = eledgedata[ni].lower_bound(fr);
		for(int i=0; i<count; ++emi, ++i) {
		  if (((*emi).second.location==(*other).second.location) && 
		      ((*emi).second.entry==(*other).second.entry) ) {
		    eledgedata[ni].erase(emi);
		    break;
		  }
		  // Still might be diagnostically useful to know if
		  // you broke out or not -- it indicates failure to
		  // find a matching intersection.  Unlikely, but
		  // very serious.
		}
	      }
	      
	      // Remove the dealt-with entries from the copy of the
	      // map, and go around again looking for another pair.
	      key_range.erase(furst);
	      key_range.erase(other);
	      done = false;
	      break;
	    }
	  }
	}
      }
      else { // Key only occurs once, move on.
	++x;
      }
    }
      

    // Now traverse the boundary of the element, adding in the area
    // contribution from all the portions that are between an exit
    // and an entry.
      
    int icount = 0;
    for(n=0;n<nn;n++) {
      icount += eledgedata[n].size();
    }

    if (icount > 0) {
      Coord starting_point;
      Coord last_exit;  // to Brooklyn
      int current_segment = 0;
      bool done=false;
      bool started=false;
      bool outside=false;
      ElEdgeMap::iterator here = eledgedata[0].begin();
      
//       // Debugging -- review all intersections.
//       std::cerr << std::endl << "All intersections: " << std::endl;
//       for(int ni = 0; ni<nn; ++ni) {
// 	for(ElEdgeMap::iterator there = eledgedata[ni].begin();
// 	    there!=eledgedata[ni].end(); ++there) {
// 	  std::cerr << "Intersection: " << (*there).second.location << 
// 	    ((*there).second.entry ? "Entry" : "Exit" ) << std::endl;
// 	}
//       }

      while(!done) {
	// If you've reached the end of a segment, switch to the
	// next segment.  If you're collecting areas, collect from
	// the last exit to the end of the segment, and set
	// last_exit to the beginning of the new segment.
	if (here == eledgedata[current_segment].end() ) {
	  current_segment = (current_segment+1)%nn;
	  here = eledgedata[current_segment].begin();
	  if (started && outside) {
	    // Segment already incremented, end_pt is end of
	    // previous segment, start of new current segment.
	    Coord end_pt = nodes[current_segment]->position();
	    accum_area += last_exit % end_pt;
	    last_exit = end_pt;
	  }
	}
	// If you've found an exit point, then first check if you've
	// started and it's the initial exit point -- if so, just
	// set "done".  Otherwise, update the last_exit and set the
	// "outside" marker to true.  If you haven't started yet,
	// then start, and set the starting point.
	else if (!(*here).second.entry) {
	  if (started && (*here).second.location==starting_point ) {
	    done = true;
	  }
	  else {
	    last_exit = (*here).second.location;
	    outside = true;
	    if (!started) {
	      started = true;
	      starting_point = (*here).second.location;
	    }
	  }
	  ++here;
	}
	// If you've found an entrance point, then add the area from
	// the last exit point to here, and set the outside flag to
	// false.
	else if ((*here).second.entry) {
	  if (started) {
	    accum_area += last_exit % (*here).second.location;
	    outside = false;
	  }
	  ++here;
	}
      }
    }
    else { 
      // If there are no intersections, we either need to add all of
      // the element boundary, or none of it.  If the area so far is
      // negative, we need all of it.
      if (accum_area < 0.0) {
	accum_area += area();
      }
    }

    // "area" now contains 2x the contribution of all loops of this
    // category to this element.
    (*result)[i] = 0.5*accum_area;
    if (accum_area != 0.0) {
      areas_zero = false;
    }

  } // End of loop over cateogries.

  // If all of the areas are zero, then the element is homogeneous.
  // Figure out which category is the right one by getting the
  // category of the central point.  The central point is used
  // because, for a homogeneous element, round-off will not put it
  // into a different category.  This may end up being slow.
  if (areas_zero) {
    Coord center;
    for(int i=0;i<nn;i++) {
      center += nodes[i]->position();
    }
    center /= nn;
    int cat = ms.category(center);
    (*result)[cat]=area();
  }
    
  return result;  // Caller must de-allocate this vector.
}

const std::vector<double> *
CSkeletonElement::categoryAreas(CMicrostructure &ms) const {
  int ncategories = ms.nCategories();
  //std::cerr << "before underlying_pixels(ms)" << std::endl;
  const std::vector<ICoord> *pixels = underlying_pixels(ms);
  //std::cerr << "after underlying_pixels(ms)" << std::endl;
  // Find the categories of the underlying pixels, and determine if
  // they're all the same.  If they are, we can skip the fractional
  // area calculation.
  if(pixels->size() == 0)
    throw ErrProgrammingError("No pixels under element!", __FILE__, __LINE__);
  std::vector<int> categories(pixels->size());
  bool uniform = true;
  categories[0] = ms.category((*pixels)[0]);
  for(int i=1; i<categories.size(); i++) {
    categories[i] = ms.category((*pixels)[i]);
    uniform = uniform && (categories[i]==categories[i-1]);
  }

  std::vector<double> *areas = new std::vector<double>(ncategories, 0.0);

  if(uniform) {
    (*areas)[categories[0]] = area();
  }
  else {
    // Fractional area calculation.
    Coord pxlsize(ms.sizeOfPixels());
    for(int i=0; i<pixels->size(); i++) {
      int category = categories[i];
      (*areas)[category] += pixel_intersection(&(*pixels)[i], &pxlsize);
    }
  }
  delete pixels;
  return areas;
}


// TODO SOON: This function is only called by the skeleton relaxation
// machinery, from skelstiffness.C and meshqualityexpansion.C.  The
// Python skeleton calls categoryAreas directly to compute its
// homogeneity.  There probably should be a C function that computes
// homogeneity directly, and it should be swigged for use in Python.
double CSkeletonElement::energyHomogeneity(CMicrostructure &ms) const {
  const std::vector<double> *ac = categoryAreas(ms);

  int N = ac->size();
  if(N <= 1) 
    {
      delete ac;
      return 0;
    }
  double e = 1.0;
  double a = area();
  double invarea = 1./a;
  double norm = N/(N-1.0);	// 1/(1-1/N)
  for(int j=0; j<N; j++) {
    e *= (1-(*ac)[j]*invarea)*norm;
  }
  delete ac;
  return e;

}

// Find the point along edge i where the pixel category in the
// Microstructure changes.  Return true iff there's just one such
// point.
bool CSkeletonElement::transitionPoint(CMicrostructure &microstructure, int i,
				       Coord *point)
  const
{
  return microstructure.transitionPoint(nodes[i]->position(),
					nodes[(i+1)%nnodes()]->position(),
					point);
}

double CSkeletonElement::edgeHomogeneity(CMicrostructure &microstructure, int i)
  const
{
  return microstructure.edgeHomogeneity(nodes[i]->position(),
					nodes[(i+1)%nnodes()]->position());
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

CSkeletonQuad::CSkeletonQuad(CSkeletonNode *n0, CSkeletonNode *n1,
			     CSkeletonNode *n2, CSkeletonNode *n3)
  : CSkeletonElement(4)
{
  nodes[0] = n0;
  nodes[1] = n1;
  nodes[2] = n2;
  nodes[3] = n3;
}

double CSkeletonQuad::shapefun(int i, const MasterCoord &mc) const {
  switch(i) {
  case 0:
    return  0.25*(mc(0) - 1.0)*(mc(1) - 1.0);
  case 1:
    return -0.25*(mc(0) + 1.0)*(mc(1) - 1.0);
  case 2:
    return  0.25*(mc(0) + 1.0)*(mc(1) + 1.0);
  case 3:
    return -0.25*(mc(0) - 1.0)*(mc(1) + 1.0);
  }
  throw ErrBadIndex(i, __FILE__, __LINE__);
}

bool CSkeletonQuad::interior(const Coord *point) const {
  //  3________2
  //   |     /|
  //   | L  / |     If a point is inside of either triangle,    
  //   |   /  |     then it is inside the quad.
  //   |  /   |
  //   | / R  |
  //   |/     |
  //   --------
  //   0      1
  Coord p0 = nodes[0]->position();
  Coord p1 = nodes[1]->position();
  Coord p2 = nodes[2]->position();
  Coord p3 = nodes[3]->position();
  double areaL = triangleArea(p0, p2, p3);
  double areaR = triangleArea(p0, p1, p2);
  if(areaR > 0.0 && areaL > 0.0)
    return insideTriangle(p0,p2,p3, *point) || insideTriangle(p0,p1,p2, *point);
  if(areaR < 0.0 && areaL > 0.0)
    // Boomerang shape.  Vertex 1 may lie inside L.  point must lie
    // within L, but not within the inverted R.
    return insideTriangle(p0,p2,p3, *point) && !insideTriangle(p0,p2,p1,*point);
  if(areaR > 0.0 && areaL < 0.0)
    return insideTriangle(p0,p1,p2, *point) && !insideTriangle(p0,p3,p2,*point);
  return false;
}

double CSkeletonQuad::area() const {
  return
    triangleArea(nodes[0]->position(), nodes[2]->position(),
		 nodes[3]->position()) +
    triangleArea(nodes[0]->position(), nodes[1]->position(),
		 nodes[2]->position());
}

bool CSkeletonQuad::illegal() const {
  Coord p0 = nodes[0]->position();
  Coord p1 = nodes[1]->position();
  Coord p2 = nodes[2]->position();
  Coord p3 = nodes[3]->position();
  return ((p1 - p0) % (p3 - p0) <= 0.0 ||
	  (p2 - p1) % (p0 - p1) <= 0.0 ||
	  (p3 - p2) % (p1 - p2) <= 0.0 ||
	  (p0 - p3) % (p2 - p3) <= 0.0);
}

const std::vector<ICoord> *
CSkeletonQuad::underlying_pixels(const CMicrostructure &microstructure) const {
  // Divide the quad into two triangles with positive area, and mark
  // the pixels under them.  This avoids the possibly complicated
  // geometry of quadrilaterals.
  const Coord c0 = nodes[0]->position();
  const Coord c1 = nodes[1]->position();
  const Coord c2 = nodes[2]->position();
  const Coord c3 = nodes[3]->position();
  CRectangle bbox(c0, c2);
  bbox.swallow(c1);
  bbox.swallow(c3);
  MarkInfo *mm = microstructure.beginMarking(bbox);
  if(triangleArea(c0, c1, c2) >= 0.0 &&
     triangleArea(c0, c2, c3) >= 0.0) {
    microstructure.markTriangle(mm, c0, c1, c2);
    microstructure.markTriangle(mm, c0, c2, c3);
  }
  else if(triangleArea(c1, c2, c3) >= 0.0 &&
          triangleArea(c2, c3, c0) >= 0.0) {
    microstructure.markTriangle(mm, c1, c2, c3);
    microstructure.markTriangle(mm, c1, c3, c0);
  }
  std::vector<ICoord> *pxls = microstructure.markedPixels(mm); // new'd vector
  microstructure.endMarking(mm);
  return pxls;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

CSkeletonTriangle::CSkeletonTriangle(CSkeletonNode *n0, CSkeletonNode *n1,
				     CSkeletonNode *n2)
  : CSkeletonElement(3)
{
  nodes[0] = n0;
  nodes[1] = n1;
  nodes[2] = n2;
}

double CSkeletonTriangle::shapefun(int i, const MasterCoord &mc) const {
  switch(i) {
  case 0:
    return mc(0);
  case 1:
    return mc(1);
  case 2:
    return 1. - mc(0) - mc(1);
  }
  throw ErrBadIndex(i, __FILE__, __LINE__);
}

bool CSkeletonTriangle::interior(const Coord *point) const {
  return insideTriangle(nodes[0]->position(), nodes[1]->position(),
			nodes[2]->position(), *point);
};

double CSkeletonTriangle::area() const {
  return triangleArea(nodes[0]->position(), nodes[1]->position(),
		      nodes[2]->position());
}

bool CSkeletonTriangle::illegal() const {
  return area() <= 0.0; 
}

const std::vector<ICoord> *
CSkeletonTriangle::underlying_pixels(const CMicrostructure &microstructure)
  const
{
  Coord p0 = nodes[0]->position();
  Coord p1 = nodes[1]->position();
  Coord p2 = nodes[2]->position();
  CRectangle bbox(p0, p1);
  bbox.swallow(p2);
  MarkInfo *mm = microstructure.beginMarking(bbox);
  microstructure.markTriangle(mm, p0, p1, p2);
  std::vector<ICoord> *pxls = microstructure.markedPixels(mm); // new'd vector
  microstructure.endMarking(mm);
  return pxls;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// OLD VERSION USING MAP JACOBIAN

// Shape energy calculation.  The shape energy is (c-1)/c, where c 
// is the ratio of the condition number squared of the jacobian of 
// the ideal->physical space map, to the condition number squared of
// the jacobian of the identity map.  For quads, the maximum condition
// number over all corners of the element is used.
//   The energy is 0 for squares and equilateral triangles, and 1
// for degenerate shapes (ones with 180-degree vertices).

// double CSkeletonTriangle::energyShape() const {
//   Coord a_vec = nodes[1]->position()-nodes[0]->position();
//   Coord b_vec = nodes[2]->position()-nodes[0]->position();

//   double num = norm2(a_vec) + norm2(b_vec) - dot(a_vec, b_vec);
//   double den = cross(a_vec, b_vec);
//   double condtn = num*num/(3.0*den*den);
//   return (condtn-1.0)/condtn;
// }

// double CSkeletonQuad::energyShape() const {
//   double eshape=0.0;
//   for(int i=0;i<4;i++) {
//     Coord a_vec = nodes[(i+1)%4]->position() - nodes[i]->position();
//     Coord b_vec = nodes[(i+3)%4]->position() - nodes[i]->position();
//     double num = norm2(a_vec)+norm2(b_vec);
//     double den = cross(a_vec, b_vec);
//     double condtn = 0.25*num*num/(den*den);
//     double corner_eshape = (condtn-1.0)/condtn;
//     if (corner_eshape > eshape)
//       eshape = corner_eshape;
//   }
//   return pow(eshape, 1.5);
// }

// // Testing new shape energy functions using STANDARD DEVIATION
// double CSkeletonTriangle::energyShape() const {
//   const double pi = acos(-1.0);
//   const double a0 = acos(cosCornerAngle(0))*180.0/pi;
//   const double a1 = acos(cosCornerAngle(1))*180.0/pi;
//   const double a2 = 180.0 - (a0+a1);
//   const double meanA = 60.0;
//   const double normA = 1.4142135623731;
//   const double angleE = sqrt(((meanA-a0)*(meanA-a0)+
// 			      (meanA-a1)*(meanA-a1)+
// 			      (meanA-a2)*(meanA-a2))/3.0)/meanA/normA;

//   const double l0 = edgeLength(0);
//   const double l1 = edgeLength(1);
//   const double l2 = edgeLength(2);
//   const double meanL = (l0+l1+l2)/3.0;
//   const double normL = 0.7071067811866;
//   const double aspectE = sqrt(((meanL-l0)*(meanL-l0)+
// 			       (meanL-l1)*(meanL-l1)+
// 			       (meanL-l2)*(meanL-l2))/3.0)/meanL/normL;

//   if (angleE >= aspectE)
//     return angleE;
//   else
//     return aspectE;
// //   const double beta = 0.5;
// //   return beta*angleE + (1.0-beta)*aspectE;
// }

// double CSkeletonQuad::energyShape() const {
//   const double pi = acos(-1.0);
//   const double a0 = acos(cosCornerAngle(0))*180.0/pi;
//   const double a1 = acos(cosCornerAngle(1))*180.0/pi;
//   const double a2 = acos(cosCornerAngle(2))*180.0/pi;
//   const double a3 = 360.0 - (a0+a1+a2);
//   const double meanA = 90.0;
//   const double angleE = sqrt(((meanA-a0)*(meanA-a0)+
// 			      (meanA-a1)*(meanA-a1)+
// 			      (meanA-a2)*(meanA-a2)+
// 			      (meanA-a3)*(meanA-a3))/4.0)/meanA;

//   const double l0 = edgeLength(0);
//   const double l1 = edgeLength(1);
//   const double l2 = edgeLength(2);
//   const double l3 = edgeLength(3);
//   const double meanL = (l0+l1+l2+l3)/4.0;
//   const double aspectE = sqrt(((meanL-l0)*(meanL-l0)+
// 			      (meanL-l1)*(meanL-l1)+
// 			      (meanL-l2)*(meanL-l2)+
// 			      (meanL-l3)*(meanL-l3))/4.0)/meanL;

//   if (angleE >= aspectE)
//     return angleE;
//   else
//     return aspectE;
// //   const double beta = 0.5;
// //   return beta*angleE + (1.0-beta)*aspectE;
// }

// END OF OLD VERSION USING MAP JACOBIAN

double max2(const double a, const double b) {
  if(a > b)
    return a;
  return b;
}

double max3(const double a, const double b, const double c) {
  if(a > b && a > c)
      return a;
  if(b > c)
    return b;
  return c;
}

double max4(const double a, const double b, const double c, const double d) {
  double maxabc = max3(a, b, c);
  if(maxabc > d)
    return maxabc;
  return d;
}

double min3(const double a, const double b, const double c) {
  if(a < b && a < c)
    return a;
  if(b < c)
    return b;
  return c;
}
double min4(const double a, const double b, const double c, const double d) {
  double minabc = min3(a, b, c);
  if(minabc < d)
    return minabc;
  return d;
}

// Testing new shape energy functions using MAX_RATIO & MAX_ANGLE
// const double degrees = 180./M_PI;

// double CSkeletonTriangle::energyShape() const {
//   const double a0 = acos(cosCornerAngle(0))*degrees;
//   const double a1 = acos(cosCornerAngle(1))*degrees;
//   const double a2 = 180.0 - (a0+a1);
//   const double l0 = edgeLength(0);
//   const double l1 = edgeLength(1);
//   const double l2 = edgeLength(2);
//   const double min_angle_energy = pow((60.0-min3(a0, a1, a2))/60.0, 2.0);
//   const double max_angle_energy = (max3(a0, a1, a2)-60.0)/120.0;
//   const double angle_energy = max2(min_angle_energy, max_angle_energy);
//   const double aspect_energy = pow(1.0 - min3(l0, l1, l2)/max3(l0, l1, l2), 2.0);
//   return max2(angle_energy, aspect_energy);
// }

// double CSkeletonQuad::energyShape() const {  
//   const double a0 = acos(cosCornerAngle(0))*degrees;
//   const double a1 = acos(cosCornerAngle(1))*degrees;
//   const double a2 = acos(cosCornerAngle(2))*degrees;
//   const double a3 = 360.0 - (a0+a1+a2);
//   const double l0 = edgeLength(0);
//   const double l1 = edgeLength(1);
//   const double l2 = edgeLength(2);
//   const double l3 = edgeLength(3);
//   const double min_angle_energy = pow((90.0-min4(a0, a1, a2, a3))/90.0, 2.0);
//   const double max_angle_energy = (max4(a0, a1, a2, a3)-90.0)/90.0;
//   const double angle_energy = max2(min_angle_energy, max_angle_energy);
//   const double aspect_energy = pow(1.0 - min4(l0, l1, l2, l3)/max4(l0, l1, l2, l3), 2.0);
//   return max2(angle_energy, aspect_energy);
// }


// STILL NEWER SHAPE ENERGY FUNCTIONS USING TRIANGLE ALTITUDES
// These should be faster to compute.

double triangleQ(const Coord &n0, const Coord &n1, const Coord &n2) {
  // The Q of a triangle is the minimum altitude, squared, over the
  // maximum side, squared.  The altitude from vertex n to side n is
  // the length of one of the other sides times the sine of the angle
  // between that side and side n.  That is,
  //        h  =  side(n+1) cross side(n)/|side(n)|

  // Sides
  Coord s0 = n1 - n2;
  Coord s1 = n2 - n0;
  Coord s2 = n0 - n1;

  double c = cross(s0, s1);
  double ll0 = s0.normsq();
  if(ll0 == 0.0) return 0.0;
  double hh0 = c*c/ll0;

  c = cross(s1, s2);
  double ll1 = s1.normsq();
  if(ll1 == 0.0) return 0.0;
  double hh1 = c*c/ll1;

  c = cross(s2, s0);
  double ll2 = s2.normsq();
  if(ll2 == 0.0) return 0.0;
  double hh2 = c*c/ll2;

  return min3(hh0, hh1, hh2)/max3(ll0, ll1, ll2);
}

double CSkeletonTriangle::energyShape() const {
  // The factor of 4/3 normalizes Q so that it's 1 for an equilateral
  // triangle.
  static const double fourthirds = 4./3.;
  double q = fourthirds*triangleQ(nodes[0]->position(), nodes[1]->position(),
				  nodes[2]->position());
  double e = 1.-q;
  return e*e*e;
}

double CSkeletonQuad::energyShape() const {
  // A quad can be divided into two triangles in two ways.  The Q of a
  // pair of triangles is the *geometric* mean of their individual Qs.
  // We use the geometric mean so that the Q of a pair is 0 if the Q
  // of either member of the pair is zero.  (The triangle Q is
  // normalized so that it's 1 for a right isosceles triangle.)  The Q
  // of the quad is Q of the pair of triangles with the smaller Q.
  // The energy of the quad is 1-Q.

  double q0 = triangleQ(nodes[1]->position(), nodes[2]->position(),
			nodes[3]->position());
  double q1 = triangleQ(nodes[2]->position(), nodes[3]->position(),
			nodes[0]->position());
  double q2 = triangleQ(nodes[3]->position(), nodes[0]->position(),
			nodes[1]->position());
  double q3 = triangleQ(nodes[0]->position(), nodes[1]->position(),
			nodes[2]->position());
  double q03 = q0*q3;
  double q12 = q1*q2;
  double q = 4*sqrt(q03 < q12 ? q03 : q12); // factor of 4 is normalization
  double e = 1.-q;
  return e*e*e;
}
