# -*- python -*-
# $RCSfile: fieldinit.py,v $
# $Revision: 1.20 $
# $Author: langer $
# $Date: 2004/12/17 00:06:10 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from SWIG.engine import field
from common import debug
from common import registeredclass
from common import strfunction
from common.IO import parameter
from common.IO import whoville
from common.IO import xmlmenudump
import engine.mesh
import string
import struct
import types

# FieldInit subclasses must be derived from FieldInit *and* from
# RegisteredClass.

class FieldInit:
    def apply(self, femesh, field):
        fniter = femesh.funcnode_iterator()
        while not fniter.end():
            node = fniter.node()
            position = node.position()
            for i in range(field.ndof()):
                field.setvalue(femesh, node, i, self.func(position, i))
            fniter.next()

# The FieldInitParameter is not a simple RegisteredParameter because
# it has to handle more than one RegisteredClass, for different
# varieties of Fields. 

class FieldInitParameter(parameter.ParameterBase):
    def __init__(self, name, value=None, default=None, tip=None):
        self.checker = parameter.TypeChecker(FieldInit)
        parameter.ParameterBase.__init__(self, name,
                                         value=value, default=default,
                                         tip=tip)
    def binaryRepr(self, datafile, value):
        return value.binaryRepr(datafile) # uses RegisteredClass.binaryRepr
    def binaryRead(self, parser):
        (classkey,) = struct.unpack('>i',
                                    parser.getBytes(struct.calcsize('>i')))
        regclass = parser.getObject(classkey)
        registry = regclass.registry
        registration = registeredclass.getRegistration(regclass, registry)
        argdict = {}
        for param in registration.params:
            argdict[param.name] = param.binaryRead(parser)
        return registration(**argdict)
    def valueDesc(self):
        return """A field initializer from one of the <link
        linkend='RegisteredClass:ScalarFieldInit'><classname>ScalarFieldInit</classname></link>
        or <link
        linkend='RegisteredClass:TwoVectorFieldInit'><classname>TwoVectorFieldInit</classname></link>
        classes."""


# Derived classes of FieldInit must provide func(position, component)
# that returns the value of a Field component at the given position.
# They must also indicate what type of field they work for by
# inserting themselves into the fieldInitDict dictionary.

fieldInitDict = {}

class ScalarFieldInit(registeredclass.RegisteredClass, FieldInit):
    registry = []

fieldInitDict['ScalarField'] = ScalarFieldInit

class TwoVectorFieldInit(registeredclass.RegisteredClass, FieldInit):
    registry = []
fieldInitDict['TwoVectorField'] = TwoVectorFieldInit

# Subclasses of this are stubs, existing mainly to populate the
# dictionary -- this class does everything.
class OtherMeshFieldInit:
    def __init__(self, mesh):
        self.mesh = mesh                # name of the other mesh
        self.meshobj = engine.mesh.meshes[self.mesh]
        self.source_femesh = self.meshobj.getObject()
    def apply(self, femesh, field):
         fniter = femesh.funcnode_iterator()
         while not fniter.end():
             node = fniter.node()
             position = node.position()

             el = self.meshobj.skeleton.enclosingElement(position)
             fe_el = self.source_femesh.getElement(el.index)
             masterpos = fe_el.to_master(position[0], position[1])
             fieldvals = fe_el.outputField(self.source_femesh, field, masterpos)
             field.setValueFromOutputValue(femesh, node, fieldvals)
             fniter.next()
    def shortrepr(self):
        return "mesh=%s" % self.mesh

# NVectorField isn't swigged yet
##class NVectorFieldInit(registeredclass.RegisteredClass, FieldInit):
##    registry = []
##fieldInitDict[field.NVectorField] = NVectorFieldInit
   


################    

class ConstScalarFieldInit(ScalarFieldInit):
    def __init__(self, value):
        self.value = value
    def func(self, position, component):
        return self.value
    def shortrepr(self):
        return str(self.value)

registeredclass.Registration("Constant",
                             ScalarFieldInit,
                             ConstScalarFieldInit,
                             0,
                             params= [parameter.FloatParameter('value', 0.0)],
                             tip="Initialize a scalar field with a constant value."
                             )

################

class ConstTwoVectorFieldInit(TwoVectorFieldInit):
    def __init__(self, cx, cy):
        self.cx = cx
        self.cy = cy
    def func(self, position, component):
        if component == 0:
            return self.cx
        return self.cy
    def shortrepr(self):
        return "cx=%s, cy=%s" % (str(self.cx), str(self.cy))

registeredclass.Registration("Constant",
                             TwoVectorFieldInit,
                             ConstTwoVectorFieldInit,
                             0,
                             params=[
    parameter.FloatParameter('cx', 0.0),
    parameter.FloatParameter('cy', 0.0)
    ],
                             tip="Initialize a two-vector field with a constant value."
                             )

################

class FuncScalarFieldInit(ScalarFieldInit):
    def __init__(self, function):
        debug.fmsg('function=', function)
        self.function = function
    def func(self, position, component):
        return self.function(position.x, position.y)
    def shortrepr(self):
        return str(self.function)

registeredclass.Registration("XYFunction",
                             ScalarFieldInit,
                             FuncScalarFieldInit,
                             1,
                             params=[
    strfunction.XYStrFunctionParameter('function', default='0.0',
                                       tip='A function of x and y')],
                             tip="Initialize a scalar field with a function of x and y."
                             )

################

class FuncTwoVectorFieldInit(TwoVectorFieldInit):
    def __init__(self, fx, fy):
        self.fx = fx
        self.fy = fy
    def func(self, position, component):
        if component == 0:
            return self.fx(position.x, position.y)
        return self.fy(position.x, position.y)
    def shortrepr(self):
        return "fx=%s, fy=%s" % (str(self.fx), str(self.fy))


registeredclass.Registration("XYFunction",
                             TwoVectorFieldInit,
                             FuncTwoVectorFieldInit,
                             1,
                             params=[
    strfunction.XYStrFunctionParameter('fx', value='0.0',
                        tip="The x component of the field as a function of x and y"),
    strfunction.XYStrFunctionParameter('fy', value='0.0',
                        tip="The y component of the field as a function of x and y"),
    ],
                             tip="Initialize a two-vector field with a function of x and y.")
                             
##################

class ScalarOtherMeshInit(OtherMeshFieldInit, ScalarFieldInit):
    pass
    
registeredclass.Registration("Other Mesh",
                             ScalarFieldInit,
                             ScalarOtherMeshInit,
                             2,
                             params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes)],
                             tip="Initialize a scalar field from the same field on another mesh.")

class TwoVectorOtherMeshInit(OtherMeshFieldInit, TwoVectorFieldInit):
    pass
            
registeredclass.Registration("Other Mesh",
                             TwoVectorFieldInit,
                             TwoVectorOtherMeshInit,
                             2,
                             params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes)],
                             tip="Initialize a two-vector field from the same field on another mesh.")

##############################################
    
# Because FieldInitParameters aren't normal RegisteredParameters, the
# classes have to be explicitly added to the xml docs.
xmlmenudump.addRegisteredClass(ScalarFieldInit)
xmlmenudump.addRegisteredClass(TwoVectorFieldInit)

