// -*- C++ -*-
// $RCSfile: skeletonrelaxationrate.C,v $
// $Revision: 1.1 $
// $Author: redwing $
// $Date: 2005/02/24 21:25:33 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/coord.h"
#include "engine/IO/propertyoutput.h"
#include "engine/element.h"
#include "engine/equation.h"
#include "engine/femesh.h"
#include "engine/flux.h"
#include "engine/material.h"
#include "engine/matvec.h"
#include "engine/cskeleton.h"
#include "engine/property/elasticity/cijkl.h"
#include "engine/property/elasticity/elasticity.h"
#include "engine/property/orientation/orientation.h"
#include "skeletonrelaxationrate.h"



SkeletonRelaxationRate::SkeletonRelaxationRate(PyObject *reg, 
			     const std::string &nm, 
			     double alpha, 
			     double beta)
  : Property(nm,reg),
    alpha_(alpha),
    beta_(beta)
{
  stress_flux=dynamic_cast<SymmetricTensorFlux*>(Flux::getFlux("Stress"));
}

void SkeletonRelaxationRate::cross_reference(Material *mat) {
  // find out which property is the elasticity
  elasticity = dynamic_cast<Elasticity*>(mat->fetchProperty("Elasticity"));
  
}

void SkeletonRelaxationRate::precompute() {
};

int SkeletonRelaxationRate::K_integration_order(int degree) const {
  return 2*degree - 1;
}


void SkeletonRelaxationRate::fluxmatrix(const FEMesh *mesh,
			       const Element *element,
			       const ElementFuncNodeIterator &nu, 
			       const Flux *flux, FluxData *fluxdata,
			       const MasterPosition &x) const
{
  if(*flux != *stress_flux) {
    throw ErrProgrammingError("Unexpected flux." __FILE__, __LINE__);
  }
  
  
}

void SkeletonRelaxationRate::fluxrhs(const FEMesh *mesh, const Element *element,
				   const Flux *flux, FluxData *fluxdata,
				   const MasterPosition &x) const {
  if(*flux!=*stress_flux) {
    throw ErrProgrammingError("Unexpected flux." __FILE__, __LINE__);
  }
  const Cijkl modulus = elasticity->cijkl(mesh, element, x);
  const CSkeletonElement * skelel = element->get_skeleton_element();
  double energyH = skelel->energyHomogeneity(*mesh->get_microstructure());
  const SymmMatrix3 S = shapetensor(element, mesh);
  for(SymTensorIterator ij; !ij.end(); ++ij) {
    double &offset_el = fluxdata->offset_element(mesh, ij); // reference!
    for(SymTensorIterator kl; !kl.end(); ++kl) {
      if(kl.diagonal()) {
	offset_el += beta_*modulus(ij,kl)*alpha_*(1.0+energyH)*energyH;
	offset_el += (1.0-beta_)*modulus(ij,kl)*alpha_*S(kl.row(),kl.col());
      }
      else {
	offset_el += 2.0*(1.0-beta_)*modulus(ij,kl)*alpha_*S(kl.row(),kl.col());
      }
    }
  }
}

SymmMatrix3 SkeletonRelaxationRate::shapetensor( const Element * element, const FEMesh *mesh) const {
  SymmMatrix3 etensor;
  Coord r_c;
  //  std::cerr << "setting shape tensor" << std::endl;
  const std::vector<Node*> & nl = element->get_nodelist();
  double inv_nnodes = 1./double(element->nnodes());
  for (int i = 0; i<element->nnodes(); i++)
    {
      Node * ni = nl[i];
      //r_c += ni->displaced_position(mesh);
      r_c += ni->position();
    }
  r_c *= inv_nnodes;

  Coord x;
  double xx = 0;
  double yy = 0;
  double xy = 0;
  for (int i = 0; i<element->nnodes(); i++)
    {
      Node * ni = nl[i];
      //x = ni->displaced_position(mesh) - r_c;
      x = ni->position() - r_c;
      xx += x(0)*x(0);
      yy += x(1)*x(1);
      xy += x(0)*x(1);
    }
  double traceI = xx + yy;
  if (traceI == 0)
    traceI = 1.0;
  etensor(0,0) = (xx -yy)/2.0;
  etensor(0,0) /= traceI;
  etensor(0,1) = xy;
  etensor(0,1) /= traceI;
  etensor(1,1) = -etensor(0,0);
  return etensor;
}

int SkeletonRelaxationRate::is_symmetric(const FEMesh* mesh) const {
  
  return 1;

}
