# -*- python -*-
# $RCSfile: snapnode.py,v $
# $Revision: 1.47 $
# $Author: langer $
# $Date: 2005/02/03 14:48:10 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from common import debug
from common import registeredclass
from common.IO import parameter
from common.IO import reporter
from common.IO import xmlmenudump
from engine import deputy
from engine import skeletonmodifier
from SWIG.common import ooferror
from SWIG.common.IO import progressbar

import random

# The SnapNodes SkeletonModifier moves element corners in a more or
# less deterministic fashion, to the points along the element's edges
# at which the pixel category changes ("transition points").  For each
# element, if the element meets a target criterion embodied in the given
# SnapNodeTargets object, SnapNodes first finds the transition points,
# and schedules the nodes to be moved by creating an appropriate
# NodeSnapper object.  Different kinds of NodeSnappers apply to different
# element and transition point topologies.  After all the NodeSnapper
# objects have been created, SnapNodes applies them to the skeleton in
# a random order, but giving precedence to ones with a lower
# "priority".  (Different kinds of NodeSnappers have different
# priorities-- edge motions are performed before single node motions.)
# Moves are accepted only if they lower the skeleton's total shape and
# homogeneity energy.

class SnapNodes(skeletonmodifier.SkeletonModifier):
    def __init__(self, targets, criterion):
        self.targets = targets  # SnapNodeTargets
        self.criterion = criterion  # Acceptance criterion

    def apply(self, oldskeleton, context):
        pBar = progressbar.getProgress()
        pBar.set_message("examining elements...    ")
        skel = oldskeleton.deputyCopy()
        skel.activate()

        # Examine elements and create NodeSnapper objects
        movedict = {}                   # dict of all moves, keyed by element
        movelists = {}         # dict of lists of all moves, keyed by priority
        elements = self.targets(context)

        # Big try-finally block to ensure that
        # self.targets.cleanSelection() gets called if something goes
        # wrong.
        try:
            nel = len(elements)
            for i in range(nel):
                element = elements[i]
                if element.homogeneity(skel) == 1.0:
                    continue  # no need to even look at it!
                if element.active(oldskeleton):
                    nnodes = element.nnodes()
                    transitionpts = [element.transitionPoint(skel, n)
                                     for n in range(nnodes)]
                    nodemotion = getNodeSnapper(element, transitionpts)
                    if nodemotion is not None:
                        movedict[element] = nodemotion
                        try:
                            movelists[nodemotion.priority].append(nodemotion)
                        except KeyError:
                            movelists[nodemotion.priority] = [nodemotion]
                if pBar.query_stop() :
                    pBar.set_failure()
                    pBar.set_message("Failed")
                    return None
                else:
                    pBar.set_progress(1.0*(i+1)/nel)
                    pBar.set_message("examining elements: %d/%d" % (i+1, nel))

            # Perform node motions in random order within their
            # priority classes
            priorities = movelists.keys()
            priorities.sort()
            movednodes = {}  # A dict to count no. of nodes moved.
            for p in priorities:
                movelist = movelists[p]
                random.shuffle(movelist)
                nmv = len(movelist)
                for i in range(nmv):
                    move = movelist[i]
                    # When a move is done, it has the option of performing
                    # other moves immediately.  This is why we have to
                    # check move.done here and why movedict is passed in
                    # to move.perform.
                    if not move.done:
                        moved = move.perform(skel, self.criterion, movedict)
                        for mvd in moved:
                            movednodes[mvd] = 1
                    if pBar.query_stop():
                        pBar.set_failure()
                        pBar.set_message("Failed")
                        return None
                    else:
                        pBar.set_progress(1.0*(i+1)/nmv)
                        pBar.set_message("snapping nodes: %d/%d" % (i+1, nmv))
            nmoved = len(movednodes)
        finally:  
            self.targets.cleanSelection()

        # Only need to clean up the skeleton if we're going to return it.
        skel.cleanUp()
        reporter.report("Snapped %d node%s." % (nmoved, "s"*(nmoved != 1)))
        return skel

                             
####################################

# SnapNodeTargets determine which elements will have their nodes
# moved (or attempted to be moved).  A SnapNodeTargets must have a
# __call__ method that takes an element and a skeleton and returns 1
# or 0, depending on whether the element should move its nodes or not.
# They are slightly different than SkelModTargets, since SnapNode doesn't
# need properCopy(), instead it calls deputyCopy().

class SnapNodeTargets(registeredclass.RegisteredClass):
    registry = []
    def cleanSelection(self):
        pass

    tip = "Which nodes to snap?"
    discussion = """<para>
    Values for the <varname>targets</varname> parameter of the <xref
    linkend='RegisteredClass:SnapNodes'/> &skel; <link
    linkend='RegisteredClass:SkeletonModifier'>modifier</link>.
    <varname>targets</varname> determines which &node; moves should be
    attempted.
    </para>"""

class SnapAll(SnapNodeTargets):
    def __call__(self, context):
        return context.getObject().element_iterator()

registeredclass.Registration(
    'All Nodes',
    SnapNodeTargets,
    SnapAll,
    0,
    tip="Snap all nodes.",
    discussion="""<para>
    All &nodes; of the &skel; are <varname>targets</varname> of <xref
    linkend='RegisteredClass:SnapNodes'/>.
    </para>""")

class SnapSelected(SnapNodeTargets):
    def __call__(self, context):
        return context.elementselection.retrieveInOrder()

registeredclass.Registration(
    'Selected Elements',
    SnapNodeTargets,
    SnapSelected,
    1,
    tip="Snap selected elements.",
    discussion="""<para>
    All &nodes; of the currently selected &elems; are
    <varname>targets</varname> for <xref
    linkend='RegisteredClass:SnapNodes'/>.
    </para>""")

class SnapHeterogenous(SnapNodeTargets):
    def __init__(self, threshold):
        self.threshold = threshold
        
    def __call__(self, context):
        skel = context.getObject()
        elements = []
        for elem in skel.element_iterator():
            if elem.homogeneity(skel) < self.threshold:
                elements.append(elem)
        return elements

registeredclass.Registration(
    'Heterogeneous Elements',
    SnapNodeTargets,
    SnapHeterogenous, 2,
    params=[
    parameter.FloatRangeParameter('threshold', (0.0, 1.0, 0.01), value=0.9,
          tip='Snap elements with homogeneity below the specified threshold.')],
    tip="Snap heterogeneous elements.",
    discussion="""<para>
    All &nodes; of the <link
    linkend='Section:Concepts:Skeleton:Homogeneity'>inhomogeneous</link>
    &elems; are <varname>targets</varname> for <xref
    linkend='RegisteredClass:SnapNodes'/>.  All &elems; whose
    homogeneity is less than the given <varname>threshold</varname>
    will be addressed.
    </para>""")
    

# The SnapNode machinery is element-based, but it's useful to be able
# to snap selected nodes.  This target does this by hacking around the
# problem -- it finds all the elements associated with the set of
# nodes, and pins all the nodes which are associated with those
# elements but which were not part of the initial set.  It then
# returns the element list to the SnapNode machinery, which is tricked
# into doing What We Want.  Afterwards, the SnapNodes.apply calls
# cleanSelection, where this object unpins all the nodes that it
# pinned.
class SnapSelectedNodes(SnapNodeTargets):
    def __init__(self):
        self.molested_nodes = []
    def __call__(self, context):
        elements = {}  # Used for uniqueness.
        all_nodes = {} # Keyed by nodes -- 0 means move, 1 means pin. 
        skel = context.getObject()
        nodes = context.nodeselection.retrieveInOrder()
        for n in nodes:
            all_nodes[n]=0 # Do not molest the nodes we like.
            
        for n in nodes:
            for e in n.neighborElements():
                try:
                    elements[e]
                except KeyError:
                    elements[e]=1
                    for x in e.nodes:
                        try:
                            all_nodes[x]
                        except KeyError:
                            all_nodes[x]=1 # Pin all extra nodes.
                            
        for (node, p) in all_nodes.items():
            if p==1:
                if not node.pinned():
                    node.setPinned(1)
                    self.molested_nodes.append(node)
            
        return elements.keys()
            

    # Use this hook to unpin the nodes we changed.
    def cleanSelection(self):
        for n in self.molested_nodes:
            n.setPinned(0)
        self.molested_nodes=[]
        

registeredclass.Registration(
    'Selected Nodes',
    SnapNodeTargets,
    SnapSelectedNodes, 3,
    tip="Snap selected nodes.",
    discussion="""<para>
    Use only the currently selected &nodes; as the
    <varname>targets</varname> for <xref
    linkend='RegisteredClass:SnapNodes'/>.
    </para>""")

###################################

registeredclass.Registration(
    'Snap Nodes',
    skeletonmodifier.SkeletonModifier,
    SnapNodes,
    ordering = 1,
    params=[parameter.RegisteredParameter('targets', SnapNodeTargets,
                                          tip='Which elements to snap.'),
            parameter.RegisteredParameter('criterion',
                                          skeletonmodifier.SkelModCriterion,
                                          tip='Acceptance criterion.')
    ],
    tip="Move nodes directly to pixel boundaries.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg-snapnodes.xml'))


####################################

# To associate the proper NodeSnapper class with a element, we compute a
# node motion signature and compare it with the signatures that each
# NodeSnapper can handle.  (Node motion signatures are similar to
# refinement signatures, but simpler.)  The signature is a tuple with
# an entry for each edge of the element.  The entry is 1 if there's a
# transition point (possible target for node motion) on the edge, and
# 0 if there's not.  To make the signature independent of the choice
# of edge number, the signature tuple is rotated so that it's
# maximized when read as a binary number.

def findSignature(transitionpoints):
    # Returns the rotation and signature.
    n = len(transitionpoints)
    sig = [pt is not None for pt in transitionpoints] # 0's and 1's
    # put in canonical order
    max = -1
    rotation = None
    for i in range(n):
        key = reduce(lambda x,y: 2*x+y, sig[i:]+sig[:i], 0)
        if key > max:
            max = key
            rotation = i
    return rotation, tuple(sig[rotation:] + sig[:rotation])

# Dictionary of NodeSnappers, keyed by signature
moverRegistry = {}              

def registerNodeSnapper(moverclass, *signatures):
    for sig in signatures:
        moverRegistry[sig] = moverclass

def getNodeSnapper(element, transitionpoints):
    rotation, sig = findSignature(transitionpoints)
    try:
        moverclass = moverRegistry[sig] # NodeSnapper class
    except KeyError:
        return
    return moverclass(element, rotation, transitionpoints) # NodeSnapper object

#####################################

class NodeSnapper:
    # NodeSnappers must have an apply() method which actually performs
    # the move if it's acceptable.  If it performs the move, it should
    # call NodeSnapper.perform() for the movers of the adjacent
    # elements.  It should return the *total* number of nodes moved.
    def __init__(self, element, rotation, transitionpoints):
        self.element = element
        self.rotation = rotation        # rotation of nodes wrt signature
        self.transitionpoints = transitionpoints
        self.done = 0
    def perform(self, skel, criterion, movedict):
        # returns no. of nodes moved
        if not self.done:
            self.done = 1
            return self.apply(skel, criterion, movedict)
        return []
    def rotate(self, vec):
        return vec[self.rotation:] + vec[:self.rotation]

###########################

class QuadEdgeSnapper(NodeSnapper):
    priority = 1

    #  3-----T1---2
    #  |      ....|  T0 and T1 are the transition points.
    #  |      ....|  Possible sets of moves are (3->T1, 0->T0)
    #  |     .....|                         and (2->T1, 1->T0)
    #  |    ......|
    #  0----T0----1

    def apply(self, skel, criterion, movedict):
        moved = []
        nodes = self.rotate(self.element.nodes)
        points = self.rotate(self.transitionpoints)
        T0 = points[0]
        T1 = points[2]
        pair0 = (nodes[0], nodes[3])
        pair1 = (nodes[1], nodes[2])
        # Candidates for the next move
        nbrelements = (self.element.getSister(skel, nodes[0], nodes[1]),
                       self.element.getSister(skel, nodes[2], nodes[3]))
        
        # Move both pairs, and accept the better one.
        change0 = deputy.DeputyProvisionalChanges()
        change0.moveNode(pair0[0], T0)
        change0.moveNode(pair0[1], T1)
        
        change1 = deputy.DeputyProvisionalChanges()
        change1.moveNode(pair1[0], T0)
        change1.moveNode(pair1[1], T1)

        bestchange = criterion([change0, change1], skel)
        if bestchange is not None:
            moved = [mn.node for mn in bestchange.movednodes]
            bestchange.accept(skel)
            # Now try to move neighbors
            for nbrelement in nbrelements:
                try:
                    mover = movedict[nbrelement]
                except KeyError:
                    pass
                else:
                    moved += mover.perform(skel, criterion, movedict)
        return moved

registerNodeSnapper(QuadEdgeSnapper, (1,0,1,0))

###########################

class TriangleEdgeSnapper(NodeSnapper):
    priority = 1  # priority is same as QuadEdgeSnapper

    #     0
    #     |\      T0 and T1 are the transition points
    #     | \     Move 0->T0 and 2->T1
    #     |  \    
    #     T0  \ 
    #     |.   \
    #     |..   \
    #     |...   \
    #     1---T1--2
    
    def apply(self, skel, criterion, movedict):
        moved = []
        nodes = self.rotate(self.element.nodes)
        points = self.rotate(self.transitionpoints)
        node0 = nodes[0]
        T0 = points[0]
        node2 = nodes[2]
        T1 = points[1]

        change = deputy.DeputyProvisionalChanges()
        change.moveNode(node0, T0)
        change.moveNode(node2, T1)

        bestchange = criterion([change], skel)
        if bestchange is not None:
            moved = [mn.node for mn in bestchange.movednodes]
            bestchange.accept(skel)
            # Now try to move neighbors
            nbrelement = self.element.getSister(skel, node0, node2)
            try:
                mover = movedict[nbrelement]
            except KeyError:  # neighbor doesn't want to move
                pass
            else:
                moved += mover.perform(skel, criterion, movedict)
        return moved

registerNodeSnapper(TriangleEdgeSnapper, (1,1,0))

###########################

class DoubleNodeSnapper(NodeSnapper):
    priority = 2

    # Move to two transition points on adjacent edges, in cases where
    # an edge move is impossible.  This can only happen on
    # quadrilaterals.

    #    2-----T1-----1
    #    |........    |   T0 and T1 are the transition points.
    #    |..........  |   
    #    |............T0  Pick the best of the following three moves:  
    #    |............|      a) 0->T0  1->T1
    #    |............|      b) 0->T0  2->T1
    #    |............|      c) 1->T0  2->T1
    #    3------------0   
    def apply(self, skel, criterion, movedict):
        moved = []
        nodes = self.rotate(self.element.nodes)
        points = self.rotate(self.transitionpoints)
        pair0 = (nodes[0], nodes[1])
        pair1 = (nodes[0], nodes[2])
        pair2 = (nodes[1], nodes[2])
        T0 = points[0]
        T1 = points[1]
        # Candidates for the next move
        nbrelements = (self.element.getSister(skel, nodes[0], nodes[1]),
                       self.element.getSister(skel, nodes[1], nodes[2]))
        
        # Move all pairs, and accept the best one.
        change0 = deputy.DeputyProvisionalChanges()
        change0.moveNode(pair0[0], T0)
        change0.moveNode(pair0[1], T1)
        
        change1 = deputy.DeputyProvisionalChanges()
        change1.moveNode(pair1[0], T0)
        change1.moveNode(pair1[1], T1)

        change2 = deputy.DeputyProvisionalChanges()
        change2.moveNode(pair2[0], T0)
        change2.moveNode(pair2[1], T1)

        bestchange = criterion([change0, change1, change2], skel)
        if bestchange is not None:
            moved = [mn.node for mn in bestchange.movednodes]
            bestchange.accept(skel)
            # Now try to move neighbors
            for nbrelement in nbrelements:
                try:
                    mover = movedict[nbrelement]
                except KeyError:
                    pass
                else:
                    moved += mover.perform(skel, criterion, movedict)
        return moved
    
registerNodeSnapper(DoubleNodeSnapper, (1,1,0,0))

###########################
                  
class SingleNodeSnapper(NodeSnapper):
    priority = 3

    #   ----1
    #       |    Move either 0 or 1 to T.
    #       |
    #       T
    #       |
    #       |
    #       |
    #   ----0  

    def apply(self, skel, criterion, movedict):
        moved = []
        nodes = self.rotate(self.element.nodes)
        points = self.rotate(self.transitionpoints)
        node0 = nodes[0]
        node1 = nodes[1]
        T = points[0]

        change0 = deputy.DeputyProvisionalChanges()
        change0.moveNode(node0, T)
        
        change1 = deputy.DeputyProvisionalChanges()
        change1.moveNode(node1, T)

        bestchange = criterion([change0, change1], skel) 
        if bestchange is not None:
            moved = [mn.node for mn in bestchange.movednodes]
            bestchange.accept(skel)
            # Now try to move neighbors
            nbrelement = self.element.getSister(skel, node0, node1)
            try:
                mover = movedict[nbrelement]
            except KeyError:  # neighbor doesn't want to move
                pass
            else:
                moved += mover.perform(skel, criterion, movedict)
        return moved

registerNodeSnapper(SingleNodeSnapper, (1,0,0), (1,0,0,0))

